﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;
using HaloMap;

namespace Gravemind
{
    public partial class MapForm : Form
    {
        #region Enums
        public enum RawState
        {
            Bitmap,
            Collision,
            Decorator,
            Particle,
            Model,
            BSP,
            Lightmap,
            Sound,
            None
        }
        #endregion

        #region Fields
        string Location;
        public Map Map;
        public RawState State = RawState.None;
        #endregion

        #region Intialize
        public MapForm(string Location)
        {
            InitializeComponent();
            this.Location = Location;
            Map = new Map(Location);
        }

        private void MapForm_Load(object sender, EventArgs e)
        {
            ListViewPanel.Visible = true;
            MetaHolder.Visible = false;
            this.Show();
            Map.Intialize(toolStripStatusLabel1, treeView1);
            this.Text = "[" + Map.Header.Scenario + "] - " + Location;
            lblMapName.Text = Map.Header.MapName;
            lblMapSize.Text = Map.Header.MapSize.ToString();
            lblMapType.Text = Map.Header.MapType.ToString();
            lblTagCount.Text = Map.Header.TagCount.ToString();
            //pictureBox1.Image = Map.LoadPreview();
            //try { pictureBox1.Image = Map.LoadPreview(); }
            //catch { }
        }
        #endregion

        #region TreeView + ListView
        private void treeView1_AfterSelect(object sender, TreeViewEventArgs e)
        {
            try
            {
                Map.UpdateProccess("Updating...");
                Map.SelectedTag = Map.FindTag(treeView1.SelectedNode.Parent.Text, treeView1.SelectedNode.Text);
                lblTagClass.Text = Map.SelectedTag.Class;
                lblTagOffset.Text = Map.SelectedTag.MetaOffset.ToString();
                lblTagSize.Text = Map.SelectedTag.MetaSize.ToString();
                listViewEx1.Clear();
                btnBitmap.Visible = false;
                btnModel.Visible = false;
                toolStripButton3.Visible = false;
                toolStripButton1.Visible = false;
                toolStripButton5.Visible = false;
                MEDL.Enabled = false;
            }
            catch (Exception)
            {
                Map.UpdateProccess("Idle...");
                lblTagClass.Text = null;
                lblTagOffset.Text = null;
                lblTagSize.Text = null;
                lblRawType.Text = null;
                lblRawOffset.Text = null;
                lblRawSize.Text = null;
                listViewEx1.Clear();
                MetaPanel.Controls.Clear();
                btnBitmap.Visible = false;
                btnModel.Visible = false;
                toolStripButton3.Visible = false;
                toolStripButton1.Visible = false;
                toolStripButton5.Visible = false;
                MEDL.Enabled = false;
                return;
            }
            if (Map.SelectedTag.Identifier == -1)
            {
                Map.UpdateProccess("Idle...");
                lblTagClass.Text = null;
                lblTagOffset.Text = null;
                lblTagSize.Text = null;
                lblRawType.Text = null;
                lblRawOffset.Text = null;
                lblRawSize.Text = null;
                listViewEx1.Clear();
                MetaPanel.Controls.Clear();
                btnBitmap.Visible = false;
                btnModel.Visible = false;
                toolStripButton3.Visible = false;
                toolStripButton1.Visible = false;
                toolStripButton5.Visible = false;
                State = RawState.None;
                MEDL.Enabled = false;
                return;
            }
            try
            {
                #region Tag Specific
                if (Map.SelectedTag.Class == "bitm")
                {
                    Map.UpdateProccess("Reading Bitmap Raw...");
                    btnBitmap.Visible = true;
                    BitmapRaw.Read(Map);
                    lblRawType.Text = "Bitmap";
                    lblRawOffset.Text = BitmapRaw.RawChunks[0].LOD1.ToString();
                    lblRawSize.Text = BitmapRaw.RawChunks[0].Size1.ToString();
                    Map.UpdateProccess("Idle...");
                }
                else if (Map.SelectedTag.Class == "DECR")
                {
                    Map.UpdateProccess("Reading Decorator Raw...");
                    toolStripButton3.Visible = true;
                    DecoratorRaw.Read(Map);
                    lblRawType.Text = "Decorator";
                    lblRawOffset.Text = DecoratorRaw.Decorator[0].Offset.ToString();
                    lblRawSize.Text = DecoratorRaw.Decorator[0].Size.ToString();
                }
                else if (Map.SelectedTag.Class == "jmad")
                {
                    Map.UpdateProccess("Reading Animation Raw...");
                    //AnimationRaw.Read(Map);
                }
                else if (Map.SelectedTag.Class == "mode")
                {
                    Map.UpdateProccess("Reading Model Raw...");
                    btnModel.Visible = true;
                    ModelRaw.Read(Map);
                    lblRawType.Text = "Model";
                    lblRawOffset.Text = ModelRaw.Model1[0].Offset.ToString();
                    lblRawSize.Text = ModelRaw.Model1[0].Size.ToString();
                }
                else if (Map.SelectedTag.Class == "PRTM")
                {
                    Map.UpdateProccess("Reading Particle Raw...");
                    ParticleRaw.Read(Map);
                    lblRawType.Text = "Particle";
                    lblRawOffset.Text = ParticleRaw.Particles[0].Offset.ToString();
                    lblRawSize.Text = ParticleRaw.Particles[0].Size.ToString();
                }
                else if (Map.SelectedTag.Class == "sbsp")
                {
                    //int Index = Map.FindBsp(Map.SelectedTag.Identifier);
                    toolStripButton5.Visible = true;
                    lblRawType.Text = "Bsp";
                    //lblRawOffset.Text = Map.Bsps[Index].Permutations[0].RawOffset.ToString();
                    //lblRawSize.Text = Map.Bsps[Index].Permutations[0].RawSize.ToString();
                }
                else if (Map.SelectedTag.Class == "ltmp")
                {
                    int Index = Map.FindLightmap(Map.SelectedTag.Identifier);
                    lblRawType.Text = "Lightmap";
                    lblRawOffset.Text = Map.Lightmaps[Index].Groups[0].Clustors[0].RawOffset.ToString();
                    lblRawSize.Text = Map.Lightmaps[Index].Groups[0].Clustors[0].RawSize.ToString();
                }
                else if (Map.SelectedTag.Class == "snd!")
                {
                    Map.UpdateProccess("Reading Sound Raw...");
                    toolStripButton1.Visible = true;
                    SoundRaw.Read(Map);
                    lblRawType.Text = "Sound";
                    lblRawOffset.Text = SoundRaw.Permutation.Choices[0].SoundChunks[0].Offset.ToString();
                    lblRawSize.Text = SoundRaw.Permutation.Choices[0].SoundChunks[0].Size.ToString();
                }
                else if (Map.SelectedTag.Class == "weat")
                {
                    Map.UpdateProccess("Reading Weather Raw...");
                    WeatherRaw.Read(Map);
                    lblRawType.Text = "Weather";
                    lblRawOffset.Text = WeatherRaw.Weather[0].Offset.ToString();
                    lblRawSize.Text = WeatherRaw.Weather[0].Size.ToString();
                }
                else
                {
                    State = RawState.None;
                    btnBitmap.Visible = false;
                    btnModel.Visible = false;
                    toolStripButton3.Visible = false;
                    toolStripButton1.Visible = false;
                }
                #endregion

                #region Raw Tools
                if (State == RawState.Bitmap)
                {
                    BitmapEditor b = (BitmapEditor)MetaPanel.Controls[0];
                    b.Update(Map);
                    return;
                }
                else if (State == RawState.Model)
                {
                    ModelEditor m = (ModelEditor)MetaPanel.Controls[0];
                    m.Update(Map);
                    return;
                }
                else if (State == RawState.Decorator)
                {
                    DecoratorEditor d = (DecoratorEditor)MetaPanel.Controls[0];
                    d.Update(Map);
                    return;
                }
                else if (State == RawState.Sound)
                {
                    SoundEditor s = (SoundEditor)MetaPanel.Controls[0];
                    s.Update(Map);
                    return;
                }
                #endregion

                #region Meta Tools
                if (referenceEditorToolStripMenuItem.Checked == true && tagReferencesToolStripMenuItem.Checked == true)
                {
                    LoadIdents();
                }
                else if (referenceEditorToolStripMenuItem.Checked == true && reflexivesToolStripMenuItem.Checked == true)
                {
                    LoadReflexives();
                }
                else if (referenceEditorToolStripMenuItem.Checked == true && scriptReferencesToolStripMenuItem.Checked == true)
                {
                    LoadScriptReferences1();
                }
                else if (metaEditorToolStripMenuItem.Checked == true)
                {
                    LoadMetaEditor();
                }
                #endregion

                Map.UpdateProccess("Idle...");
            }
            catch (Exception)
            {
                lblRawType.Text = null;
                lblRawOffset.Text = null;
                lblRawSize.Text = null;
                btnBitmap.Visible = false;
                toolStripButton3.Visible = false;
                toolStripButton1.Visible = false;
                toolStripButton5.Visible = false;
                btnModel.Visible = false;
                MEDL.Enabled = false;
                Map.UpdateProccess("Idle...");
                return;
            }
        }
        #endregion

        #region Script References
        private void scriptReferencesToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = true;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            LoadScriptReferences();
        }

        public void LoadScriptReferences()
        {
            toolStripStatusLabel1.Text = "Loading Script References...";
            Application.DoEvents();
            listViewEx1.Clear();
            listViewEx1.Columns.Add("Script Reference", 400);
            listViewEx1.Columns.Add("Length", 97);
            for (int i = 0; i < Map.ScriptReferences.Count; i++)
            {
                ListViewItem ls = new ListViewItem(Map.ScriptReferences[i].Reference);
                ls.SubItems.Add(Map.ScriptReferences[i].Reference.Length.ToString());
                listViewEx1.Items.Add(ls);
            }
            toolStripStatusLabel1.Text = "Idle...";
        }
        #endregion

        #region References
        private void tagReferencesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = true;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;
            LoadIdents();
        }

        public void LoadIdents()
        {
            listViewEx1.Clear();
            Map.UpdateProccess("Reading Idents...");
            PluginReader.LoadIdents(Map);

            listViewEx1.Columns.Add("Description", 200);
            listViewEx1.Columns.Add("Class", 70);
            listViewEx1.Columns.Add("Offset", 81);
            listViewEx1.Columns.Add("Path", 400);
            foreach (TagReference t in Map.SelectedTag.Idents)
            {
                ListViewItem ls = new ListViewItem(t.Description);
                ls.SubItems.Add(t.Class);
                ls.SubItems.Add(t.Offset.ToString());
                ls.SubItems.Add(t.Tag.Path);
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }

        private void reflexivesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = true;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;
            LoadReflexives();
        }

        public void LoadReflexives()
        {
            listViewEx1.Clear();
            Map.UpdateProccess("Reading Reflexives...");
            PluginReader.LoadReflexives(Map);

            listViewEx1.Columns.Add("Description", 200);
            listViewEx1.Columns.Add("Offset", 81);
            listViewEx1.Columns.Add("Chunk Count", 95);
            listViewEx1.Columns.Add("Chunk Size", 95);
            listViewEx1.Columns.Add("Translation", 95);
            foreach (Reflexive r in Map.SelectedTag.Reflexives)
            {
                ListViewItem ls = new ListViewItem(r.Description);
                ls.SubItems.Add(r.Offset.ToString());
                ls.SubItems.Add(r.ChunkCount.ToString());
                ls.SubItems.Add(r.ChunkSize.ToString());
                ls.SubItems.Add(r.Translation.ToString());
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }

        private void scriptReferencesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = true;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;
            LoadScriptReferences1();
        }

        public void LoadScriptReferences1()
        {
            listViewEx1.Clear();
            Map.UpdateProccess("Reading Script References...");
            PluginReader.LoadScriptReferences(Map);

            listViewEx1.Columns.Add("Description", 200);
            listViewEx1.Columns.Add("Offset", 81);
            listViewEx1.Columns.Add("Reference", 211);
            foreach (ScriptReference s in Map.SelectedTag.ScriptReferences)
            {
                ListViewItem ls = new ListViewItem(s.Description);
                ls.SubItems.Add(s.Offset.ToString());
                ls.SubItems.Add(s.Reference);
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }

        private void referenceEditorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            metaEditorToolStripMenuItem.Checked = false;
            referenceEditorToolStripMenuItem.Checked = true;
        }
        #endregion

        #region Meta
        private void metaToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (Map.SelectedTag != null)
            {
                SaveFileDialog sfd = new SaveFileDialog();
                if (sfd.ShowDialog() == DialogResult.OK)
                {
                    Map.UpdateProccess("Decompiling Meta...");
                    Map.TagHandler.ExtractMeta(Map, Map.SelectedTag, sfd.FileName);
                    Map.UpdateProccess("Idle...");
                    MessageBox.Show("Done!");
                }
            }
        }

        private void tagToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            if (Map.SelectedTag != null)
            {
                FolderBrowserDialog ofd = new FolderBrowserDialog();
                if (ofd.ShowDialog() == DialogResult.OK)
                {
                    Map.UpdateProccess("Decompiling " + Map.SelectedTag.Path + "...");
                    Map.TagHandler.RecrusiveExtract(Map, ofd.SelectedPath);
                    Map.UpdateProccess("Idle...");
                    MessageBox.Show("Done!");
                }
            }
        }

        private void metaToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            if (Map.SelectedTag != null)
            {
                OpenFileDialog ofd = new OpenFileDialog();
                string Class = Map.SelectedTag.Class.Replace('*', '!');
                Class = Class.Replace('<', '_');
                Class = Class.Replace('>', '_');
                ofd.Filter = "Halo 2 Vista Meta (*." + Class + "|*." + Class;
                if (ofd.ShowDialog() == DialogResult.OK)
                {
                    Map.UpdateProccess("Compiling Meta...");
                    Map.TagHandler.InjectMeta(Map, Map.SelectedTag, ofd.FileName);
                    Map.UpdateProccess("Idle...");
                    MessageBox.Show("Done!");
                }
            }
        }

        private void tagToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                Map.UpdateProccess("Building...");
                this.Enabled = false;
                Map.TagHandler.Build(Map, ofd.FileName);
                Map.UpdateProccess("Refreshing...");
                lblMapSize.Text = Map.Header.MapSize.ToString();
                lblTagCount.Text = Map.Header.TagCount.ToString();
                lblTagClass.Text = null;
                lblTagOffset.Text = null;
                lblTagSize.Text = null;
                lblRawType.Text = null;
                lblRawOffset.Text = null;
                lblRawSize.Text = null;
                listViewEx1.Clear();
                listViewEx1.Visible = true;
                MetaHolder.Hide();
                MetaPanel.Controls.Clear();
                treeView1.Nodes.Clear();
                Map.Clear();
                Map.Intialize(toolStripStatusLabel1, treeView1);
                this.Enabled = true;
                Map.UpdateProccess("Idle...");
                MessageBox.Show("Done!");
            }
        }
        #endregion

        #region Unicode
        private void englishToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = true;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;

            listViewEx1.Clear();
            Map.UpdateProccess("Loading Unicode...");
            listViewEx1.Columns.Add("Unicode String", 500);
            listViewEx1.Columns.Add("Length", 72);
            foreach (UnicodeEntry u in Map.Unicode.Tables[0].UnicodeStrings)
            {
                ListViewItem ls = new ListViewItem(u.Unicode);
                ls.SubItems.Add(u.Length.ToString());
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }

        private void germanToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = true;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;

            listViewEx1.Clear();
            Map.UpdateProccess("Loading Unicode...");
            listViewEx1.Columns.Add("Unicode String", 500);
            listViewEx1.Columns.Add("Length", 72);
            foreach (UnicodeEntry u in Map.Unicode.Tables[2].UnicodeStrings)
            {
                ListViewItem ls = new ListViewItem(u.Unicode);
                ls.SubItems.Add(u.Length.ToString());
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }

        private void frenchToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = true;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;

            listViewEx1.Clear();
            Map.UpdateProccess("Loading Unicode...");
            listViewEx1.Columns.Add("Unicode String", 500);
            listViewEx1.Columns.Add("Length", 72);
            foreach (UnicodeEntry u in Map.Unicode.Tables[3].UnicodeStrings)
            {
                ListViewItem ls = new ListViewItem(u.Unicode);
                ls.SubItems.Add(u.Length.ToString());
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }

        private void spanishToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = true;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;

            listViewEx1.Clear();
            Map.UpdateProccess("Loading Unicode...");
            listViewEx1.Columns.Add("Unicode String", 500);
            listViewEx1.Columns.Add("Length", 72);
            foreach (UnicodeEntry u in Map.Unicode.Tables[4].UnicodeStrings)
            {
                ListViewItem ls = new ListViewItem(u.Unicode);
                ls.SubItems.Add(u.Length.ToString());
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }

        private void italianToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = true;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;

            listViewEx1.Clear();
            Map.UpdateProccess("Loading Unicode...");
            listViewEx1.Columns.Add("Unicode String", 500);
            listViewEx1.Columns.Add("Length", 72);
            foreach (UnicodeEntry u in Map.Unicode.Tables[5].UnicodeStrings)
            {
                ListViewItem ls = new ListViewItem(u.Unicode);
                ls.SubItems.Add(u.Length.ToString());
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }

        private void koreanToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = true;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;

            listViewEx1.Clear();
            Map.UpdateProccess("Loading Unicode...");
            listViewEx1.Columns.Add("Unicode String", 500);
            listViewEx1.Columns.Add("Length", 72);
            foreach (UnicodeEntry u in Map.Unicode.Tables[6].UnicodeStrings)
            {
                ListViewItem ls = new ListViewItem(u.Unicode);
                ls.SubItems.Add(u.Length.ToString());
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }

        private void chineseToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = true;
            japaneseToolStripMenuItem.Checked = false;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;

            listViewEx1.Clear();
            Map.UpdateProccess("Loading Unicode...");
            listViewEx1.Columns.Add("Unicode String", 500);
            listViewEx1.Columns.Add("Length", 72);
            foreach (UnicodeEntry u in Map.Unicode.Tables[7].UnicodeStrings)
            {
                ListViewItem ls = new ListViewItem(u.Unicode);
                ls.SubItems.Add(u.Length.ToString());
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }

        private void japaneseToolStripMenuItem_Click(object sender, EventArgs e)
        {
            metaEditorToolStripMenuItem.Checked = false;
            tagReferencesToolStripMenuItem.Checked = false;
            reflexivesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem.Checked = false;
            scriptReferencesToolStripMenuItem1.Checked = false;
            englishToolStripMenuItem.Checked = false;
            germanToolStripMenuItem.Checked = false;
            frenchToolStripMenuItem.Checked = false;
            spanishToolStripMenuItem.Checked = false;
            italianToolStripMenuItem.Checked = false;
            koreanToolStripMenuItem.Checked = false;
            chineseToolStripMenuItem.Checked = false;
            japaneseToolStripMenuItem.Checked = true;
            listViewEx1.Visible = true;
            MetaHolder.Visible = false;
            referenceEditorToolStripMenuItem.Checked = true;

            listViewEx1.Clear();
            Map.UpdateProccess("Loading Unicode...");
            listViewEx1.Columns.Add("Unicode String", 500);
            listViewEx1.Columns.Add("Length", 72);
            foreach (UnicodeEntry u in Map.Unicode.Tables[1].UnicodeStrings)
            {
                ListViewItem ls = new ListViewItem(u.Unicode);
                ls.SubItems.Add(u.Length.ToString());
                listViewEx1.Items.Add(ls);
            }
            Map.UpdateProccess("Idle...");
        }
        #endregion

        #region Reference Swapping
        private void swapToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (tagReferencesToolStripMenuItem.Checked == true)
            {
                try
                {
                    int[] Index = { listViewEx1.SelectedItems[0].Index };
                    TagSwapper s = new TagSwapper(this, Index);
                    s.ShowInTaskbar = false;
                    s.Show();
                }
                catch {}
            }
            else if (scriptReferencesToolStripMenuItem.Checked == true)
            {
                try
                {
                    ScriptReferenceSwapper s = new ScriptReferenceSwapper(this, listViewEx1.SelectedItems[0].Index);
                    s.ShowInTaskbar = false;
                    s.Show();
                }
                catch { }
            }
        }

        private void nullToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (tagReferencesToolStripMenuItem.Checked)
            {
                foreach (ListViewItem ls in listViewEx1.SelectedItems)
                {
                    TagReference t = new TagReference();
                    t.Class = ls.SubItems[1].Text;
                    t.Offset = Convert.ToInt32(ls.SubItems[2].Text);
                    t.Tag = new Tag();
                    t.Tag.Path = ls.SubItems[3].Text;
                    t.Null(Map);
                }
                LoadIdents();
                MessageBox.Show("Done!");
            }
        }

        private void swapToToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (tagReferencesToolStripMenuItem.Checked)
            {
                List<int> Index = new List<int>();
                foreach (ListViewItem ls in listViewEx1.SelectedItems)
                {
                    Index.Add(ls.Index);
                }

                TagSwapper swap = new TagSwapper(this, Index.ToArray());
                swap.ShowInTaskbar = false;
                swap.Show();
            }
        }
        #endregion

        #region Misc
        private void gotoToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                foreach (TreeNode n in treeView1.Nodes[0].Nodes)
                {
                    if (n.Text == listViewEx1.SelectedItems[0].SubItems[1].Text)
                    {
                        foreach (TreeNode nn in n.Nodes)
                        {
                            if (nn.Text == listViewEx1.SelectedItems[0].SubItems[3].Text)
                            {
                                treeView1.SelectedNode = nn;
                            }
                        }
                    }
                }
            }
            catch (Exception) { return; }
        }
        #endregion

        #region Meta Editor
        private void metaEditorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LoadMetaEditor();   
        }

        public void LoadMetaEditor()
        {
            MetaPanel.Controls.Clear();
            listViewEx1.Visible = false;
            MetaHolder.Visible = true;
            MEDL.Enabled = true;
            metaEditorToolStripMenuItem.Checked = true;
            referenceEditorToolStripMenuItem.Checked = false;

            label11.Visible = true;
            Map.UpdateProccess("Reading Meta...");
            PluginReader.Read(Map, ref MetaPanel);
            Map.UpdateProccess("Updating...");
            lblMEClass.Text = PluginReader.Class;
            lblAuthor.Text = PluginReader.Author;
            if (PluginReader.Revisions.Count >= 1)
            {
                lblRevision.Text = PluginReader.Revisions[0];
            }
            Map.UpdateProccess("Idle...");
            label11.Hide();
            MetaPanel.Visible = true;
        }

        private void button2_Click(object sender, EventArgs e)
        {
            PluginReader.Save(Map, ref MetaPanel);
            MessageBox.Show("Done!");
        }
        #endregion

        #region Resign
        public void Resign()
        {
            Map.UpdateProccess("Resigning...");
            Map.Resign();
            Map.UpdateProccess("Idle...");
            MessageBox.Show("Done!");
        }
        #endregion

        #region Tools
        //private void viewToolStripMenuItem_Click(object sender, EventArgs e)
        //{
        //    if (Map.SelectedTag.Class == "coll")
        //    {
        //        Globals.Initialize();
        //        CollisionViewer CollViewer = new CollisionViewer(Map);
        //    }
        //    else if (Map.SelectedTag.Class == "matg" && Map.Header.MapType == MapType.Mainmenu)
        //    {
        //        MetaPanel.Controls.Clear();
        //        listViewEx1.Visible = false;
        //        MetaHolder.Visible = true;

        //        MainMenu.Read(Map);
        //        Slot s = new Slot(Map, MainMenu.SPLevels, MainMenu.MPLevels);
        //        s.Dock = DockStyle.Top;
        //        MetaPanel.Controls.Add(s);
        //    }
        //}
        #endregion

        #region Dispose
        private void MapForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            Map.Dispose();
        }
        #endregion

        #region Raw Data
        private void btnBitmap_Click(object sender, EventArgs e)
        {
            State = RawState.Bitmap;
            MetaPanel.Controls.Clear();
            listViewEx1.Visible = false;
            MetaHolder.Visible = true;
            MetaPanel.Controls.Add(new BitmapEditor(Map));
            MetaPanel.Visible = true;
        }

        private void btnModel_Click(object sender, EventArgs e)
        {
            State = RawState.Model;
            MetaPanel.Controls.Clear();
            listViewEx1.Visible = false;
            MetaHolder.Visible = true;
            MetaPanel.Visible = true;
            MetaPanel.Controls.Add(new ModelEditor(Map));
        }

        private void toolStripButton3_Click(object sender, EventArgs e)
        {
            State = RawState.Decorator;
            MetaPanel.Controls.Clear();
            listViewEx1.Visible = false;
            MetaHolder.Visible = true;
            MetaPanel.Visible = true;
            MetaPanel.Controls.Add(new DecoratorEditor(Map));
        }

        private void toolStripButton1_Click(object sender, EventArgs e)
        {
            State = RawState.Sound;
            MetaPanel.Controls.Clear();
            listViewEx1.Visible = false;
            MetaHolder.Visible = true;
            MetaPanel.Visible = true;
            MetaPanel.Controls.Add(new SoundEditor(Map));
        }
        #endregion

        #region ChunkClonning
        private void chunkClonnerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (Map.SelectedTag != null && Map.SelectedTag.Identifier != -1)
            {
                ChunkClonner cc = new ChunkClonner(this, Map);
                cc.ShowInTaskbar = false;
                cc.Show();
            }
        }
        #endregion

        #region Real Time Halo
        private void pokeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            PluginReader.Poke(Map, ref MetaPanel);
        }
        #endregion

        private void mapToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog fbd = new FolderBrowserDialog();
            if (fbd.ShowDialog() == DialogResult.OK)
            {
                Map.DecompileMap(fbd.SelectedPath);
                MessageBox.Show("Done!");
            }
        }

        private void toolStripButton5_Click(object sender, EventArgs e)
        {
            Bsp b = new Bsp();
            b.Read(Map, Map.SelectedTag);
            State = RawState.BSP;
            BSPViewer BSP = new BSPViewer(Map, b);
        }

        private void bSPToolStripMenuItem_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog fbd = new FolderBrowserDialog();
            if (fbd.ShowDialog() == DialogResult.OK)
            {
                Map.DecompileBSP(fbd.SelectedPath);
                MessageBox.Show("Done!");
            }
        }
    }
}
