﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.VisualBasic;

namespace HaloMap
{
    public enum StringType
    {
        String32,
        String64,
        Unicode64,
        Unicode256
    }

    public class TagReference
    {
        #region Fields
        public int Offset;
        public string Class;
        public int Ident;
        public string Description;
        public Tag Tag;
        bool LoneID = false;
        #endregion

        #region Methods
        public void Read(Map map, int Offset)
        {
            this.Offset = Offset;
            Read(map);
        }
        public void Read(Map map)
        {
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + Offset;
            byte[] d = map.br.ReadBytes(4);
            for (int i = 0; i < 4; i++)
            {
                Class += Convert.ToChar(d[i]);
            }
            Class = Strings.StrReverse(Class);
            Ident = map.br.ReadInt32();
            for (int i = 0; i < map.Tags.Count; i++)
            {
                if (map.Tags[i].Class == Class && (map.Tags[i].Identifier == Ident || Ident == -1))
                {
                    Tag = (Tag)map.Tags[i].Clone();
                    Tag.Path = Ident == -1 ? "Null" : Tag.Path;
                    return;
                }
                else if (map.Tags[i].Identifier == Ident)
                {
                    LoneID = true;
                    Tag =(Tag)map.Tags[i].Clone();
                    Class = Tag.Class;
                    return;
                }
            }
            LoneID = true;
            Tag = new Tag();
            Tag.Identifier = -1;
            Class = "Null";
            Tag.Path = "Null";
        }
        public void Read(Map map, int Offset, bool b)
        {
            map.br.BaseStream.Position = Offset;
            Read(map, b);
        }
        public void Read(Map map, bool b)
        {
            Offset = (int)map.br.BaseStream.Position;
            byte[] d = map.br.ReadBytes(4);
            for (int i = 0; i < 4; i++)
            {
                Class += Convert.ToChar(d[i]);
            }
            Class = Strings.StrReverse(Class);
            Ident = map.br.ReadInt32();
            for (int i = 0; i < map.Tags.Count; i++)
            {
                if (map.Tags[i].Class == Class && (map.Tags[i].Identifier == Ident || Ident == -1))
                {
                    Tag = (Tag)map.Tags[i].Clone();
                    Tag.Path = Ident == -1 ? "Null" : Tag.Path;
                    return;
                }
                else if (map.Tags[i].Identifier == Ident)
                {
                    LoneID = true;
                    Tag = (Tag)map.Tags[i].Clone();//done
                    Class = Tag.Class;
                    return;
                }
            }
            LoneID = true;
            Tag = new Tag();
            Tag.Identifier = -1;
            Class = "Null";
            Tag.Path = "Null";
        }

        public void Null(Map map)
        {
            map.bw.BaseStream.Position = map.SelectedTag.MetaOffset + Offset + 4;
            map.bw.Write((int)-1);
        }

        public void Write(Map map, int Offset)
        {
            this.Offset = Offset;
            Write(map);
        }
        public void Write(Map map)
        {
            map.bw.BaseStream.Position = map.SelectedTag.MetaOffset + Offset;
            if (LoneID == true)
            {
                map.bw.BaseStream.Position += 4;
            }
            else
            {
                map.bw.Write(Strings.StrReverse(Class).ToCharArray());
            }
            if (Tag.Path == "Null")
            {
                map.bw.Write((int)-1);
            }
            else
            {
                map.bw.Write(Tag.Identifier);
            }
        }
        public void Write(Map map, bool b)
        {
            map.bw.BaseStream.Position = Offset;
            if (LoneID == true)
            {
                map.bw.BaseStream.Position += 4;
            }
            else
            {
                map.bw.Write(Strings.StrReverse(Class).ToCharArray());
            }
            if (Tag.Path == "Null")
            {
                map.bw.Write((int)-1);
            }
            else
            {
                map.bw.Write(Tag.Identifier);
            }
        }
        public void Write(Map map, int Offset, bool b)
        {
            map.bw.BaseStream.Position = Offset;
            if (LoneID == true)
            {
                map.bw.BaseStream.Position += 4;
            }
            else
            {
                map.bw.Write(Strings.StrReverse(Class).ToCharArray());
            }
            if (Tag.Path == "Null")
            {
                map.bw.Write((int)-1);
            }
            else
            {
                map.bw.Write(Tag.Identifier);
            }
        }
        #endregion
    }

    public class Reflexive
    {
        #region Fields
        public int Offset;
        public string Description;
        public int ChunkCount;
        public int ChunkSize;
        public int RawTranslation;
        public int Translation;
        public byte[] Data;
        public List<Reflexive> Chunks = new List<Reflexive>();
        #endregion

        #region Methods
        public void Read(Map map)
        {
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + Offset;
            ChunkCount = map.br.ReadInt32();
            RawTranslation = map.br.ReadInt32();
            Translation = RawTranslation - map.Index.SecondaryMagic;
        }

        public void Read(Map map, bool b)
        {
            Offset = (int)map.br.BaseStream.Position;
            ChunkCount = map.br.ReadInt32();
            RawTranslation = map.br.ReadInt32();
            Translation = RawTranslation - map.Index.SecondaryMagic;
        }

        public void Read(Map map, int Offset)
        {
            this.Offset = Offset;
            Read(map);
        }

        public void Read(Map map, int Offset, bool b)
        {
            map.br.BaseStream.Position = Offset;
            Read(map, b);
        }

        public void Write(Map map, int Offset)
        {
            this.Offset = Offset;
            Write(map);
        }
        public void Write(Map map)
        {
            map.bw.BaseStream.Position = map.SelectedTag.MetaOffset + Offset;
            map.bw.Write(ChunkCount);
            map.bw.Write(RawTranslation);
        }
        public void Write(Map map, bool b)
        {
            map.bw.BaseStream.Position = Offset;
            map.bw.Write(ChunkCount);
            map.bw.Write((Translation + map.Index.SecondaryMagic));
        }
        public void Fix(Map map, int Magic)
        {
            map.bw.BaseStream.Position = Offset;
            map.bw.Write(ChunkCount);
            map.bw.Write((Translation + Magic));
        }
        #endregion
    }
}
