﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace HaloMap
{
    public partial class ModelEditor : UserControl
    {
        public Map map;
        public Model mode;
        public Device Dev;
        public Camera Cam;
        public Material m;
        public VertexBuffer vb;

        public ModelEditor(Map map)
        {
            InitializeComponent();
            this.map = map;
            map.UpdateProccess("Loading Model Raw...");
            mode = new Model(map);
        }

        private void ModelEditor_Load(object sender, EventArgs e)
        {
            lblOffset.Text = mode.Model1[0].RawOffset.ToString();
            lblSize.Text = mode.Model1[0].RawSize.ToString();
            lblLocation.Text = mode.Model1[0].Location.ToString();
            lblVerts.Text = mode.Model1[0].VertexCount.ToString();
            lblFaces.Text = mode.Model1[0].FaceCount.ToString();
            map.UpdateProccess("Idle...");
            Initialize();
        }

        public void Update(Map map)
        {
            this.map = map;
            map.UpdateProccess("Loading Model Raw...");
            mode = new Model(map);
            map.UpdateProccess("Idle...");
            Initialize();
        }

        public void Initialize()
        {
            using (ModelEditor frm = (ModelEditor)this)
            {
                //Initlialize Graphics
                frm.InitGraphics();

                //Update Form
                Cam = new Camera(frm.pnlModel);
                frm.Show();
                frm.Focus();

                //Render
                while (frm.Created)
                {
                    //try
                    //{
                        Render();
                        Application.DoEvents();
                    //}
                    //catch { break; }
                }
            }
        }

        public void InitGraphics()
        {
            this.Show();
            this.Focus();
            PresentParameters PParms = new PresentParameters();
            PParms.Windowed = true;
            PParms.PresentationInterval = PresentInterval.Default;
            PParms.FullScreenRefreshRateInHz = PresentParameters.DefaultPresentRate;

            PParms.SwapEffect = SwapEffect.Discard;
            PParms.EnableAutoDepthStencil = true;
            PParms.AutoDepthStencilFormat = DepthFormat.D16;
            Dev = new Device(0, DeviceType.Hardware, this.pnlModel, CreateFlags.SoftwareVertexProcessing, PParms);
            Dev.DeviceReset += new EventHandler(OnResetDevice);
            OnResetDevice(Dev, null);
            SetUp();

            m = new Material();
            m.Ambient = Color.White;
            m.Diffuse = Color.White;
            m.Specular = Color.White;
        }

        public void Render()
        {
            //Clear Screen
            Dev.Clear(ClearFlags.ZBuffer | ClearFlags.Target, Color.Black, 1f, 0);

            //Begine Scene
            Dev.BeginScene();

            Cam.move();
            SetupMatrices();
            Dev.Material = m;
            //mode.DrawMeshes(ref Dev);
            Dev.RenderState.FillMode = FillMode.Solid;
            Dev.SetStreamSource(0, vb, 0);
            Dev.VertexFormat = VertexFormats.PositionNormal;
            Dev.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 0x62);
            Dev.RenderState.AlphaTestEnable = false;
            Dev.RenderState.AlphaBlendEnable = false;

            //Present
            Dev.EndScene();
            Dev.Present();
        }

        public void SetUp()
        {
            int Vert = 0;
            for (int i = 0; i < mode.Model1.Count; i++)
                Vert += mode.Model1[i].RawData.Vertices.Verts.Count;
            vb = new VertexBuffer(typeof(CustomVertex.PositionNormalTextured), Vert, Dev, Usage.WriteOnly, VertexFormats.PositionNormal, Pool.Default);
            CustomVertex.PositionNormalTextured[] Data = (CustomVertex.PositionNormalTextured[])vb.Lock(0, LockFlags.None);
            for (int x = 0; x < mode.Model1.Count; x++)
            {
                for (int i = 0; i < mode.Model1[x].RawData.Vertices.Verts.Count; i++)
                {
                    try
                    {
                        Data[i].X = mode.Model1[x].RawData.Vertices.Verts[i].X;
                        Data[i].Y = mode.Model1[x].RawData.Vertices.Verts[i].Y;
                        Data[i].Z = mode.Model1[x].RawData.Vertices.Verts[i].Z;
                        Data[i].Nx = mode.Model1[x].RawData.Normals.Normals[i].X;
                        Data[i].Ny = mode.Model1[x].RawData.Normals.Normals[i].Y;
                        Data[i].Nz = mode.Model1[x].RawData.Normals.Normals[i].Z;
                        Data[i].Tu = mode.Model1[x].RawData.UVData.UVs[i].X;
                        Data[i].Tv = mode.Model1[x].RawData.UVData.UVs[i].Y;
                    }
                    catch { }
                }
            }
            vb.Unlock();
        }

        public void SetupMatrices()
        {
            Dev.Transform.World = Matrix.Identity;
            Dev.Transform.View = Matrix.LookAtRH(Cam.Position, mode.Model1[0].RawData.Vertices.Verts[0], Cam.UpVector);
            Dev.Transform.Projection = Matrix.PerspectiveFovRH((float)Math.PI / 4.0f, 1.0f, 0.1f, 10000.0f);
        }

        public void OnResetDevice(object sender, EventArgs e)
        {
            Device dev = (Device)sender;
            // Turn off culling, so we see the front and back of the triangle
            dev.RenderState.CullMode = Cull.None;
            // Turn off D3D lighting
            dev.RenderState.Lighting = true;
            // Turn on the ZBuffer
            dev.RenderState.ZBufferEnable = true;
            dev.RenderState.FillMode = FillMode.Solid;
            dev.RenderState.SourceBlend = Blend.SourceAlpha;
            dev.RenderState.DestinationBlend = Blend.InvSourceAlpha;
            dev.RenderState.Ambient = Color.White;
            Dev.SamplerState[0].MinFilter = TextureFilter.Linear;
            Dev.SamplerState[0].MagFilter = TextureFilter.Linear;
            Dev.SamplerState[0].MipFilter = TextureFilter.Linear;
            Dev.SamplerState[1].MinFilter = TextureFilter.Linear;
            Dev.SamplerState[1].MagFilter = TextureFilter.Linear;
            Dev.SamplerState[1].MipFilter = TextureFilter.Linear;
        }

        private void pnlModel_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                Cam.change(e.X, e.Y);
            }
        }

        private void pnlModel_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                Cam.change(e.X, e.Y);
            }
        }

        private void button1_Click_1(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                StreamWriter sw = new StreamWriter(sfd.FileName);
                sw.WriteLine("#--------------------------------------------");
                sw.WriteLine("# Halo 2 Vista Model Extracted with Gravemind");
                sw.WriteLine("#--------------------------------------------");

                //Vertices
                for (int x = 0; x < mode.Model1.Count; x++)
                {
                    HaloMap.Models.VerticeBlock v = mode.Model1[x].RawData.Vertices;
                    for (int i = 0; i < v.Verts.Count; i++)
                    {
                        sw.WriteLine("v " + v.Verts[i].X.ToString() + " " + v.Verts[i].Y.ToString() + " " + v.Verts[i].Z.ToString());
                    }
                }

                //UVs
                for (int x = 0; x < mode.Model1.Count; x++)
                {
                    HaloMap.Models.UVData u = mode.Model1[x].RawData.UVData;
                    for (int i = 0; i < u.UVs.Count; i++)
                    {
                        sw.WriteLine("vt " + u.UVs[i].X.ToString() + " " + u.UVs[i].Y.ToString());
                    }
                }

                //Normals
                for (int x = 0; x < mode.Model1.Count; x++)
                {
                    HaloMap.Models.NormalBlock n = mode.Model1[x].RawData.Normals;
                    for (int i = 0; i < n.Normals.Count; i++)
                    {
                        sw.WriteLine("vn " + n.Normals[i].X.ToString() + " " + n.Normals[i].Y.ToString() + " " + n.Normals[i].Z.ToString());
                    }
                }

                //Faces - Not Working
                for (int x = 0; x < mode.Model1.Count; x++)
                {
                    HaloMap.Models.MeshInfoBlock m = mode.Model1[x].RawData.MeshInfo;
                    HaloMap.Models.IndiceBlock ib = mode.Model1[x].RawData.Indices;
                    for (int i = 0; i < ib.Indice.Count; i++)
                    {
                        sw.WriteLine("f " + (ib.Indice[i] + 1) + "/" + (ib.Indice[x] + 1) + "/" + (ib.Indice[x] + 1)
                            + "/" + (ib.Indice[x + 1] + 1) + "/" + (ib.Indice[x + 1] + 1) + "/" + (ib.Indice[x + 1] + 1)
                            + "/" + (ib.Indice[x + 2] + 1) + "/" + (ib.Indice[x + 2] + 1) + "/" + (ib.Indice[x + 2] + 1));
                    }
                }

                //Done
                sw.Close();
                MessageBox.Show("Done!");
            }
        }
    }
}
