﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Xml;

namespace HaloMap
{
    public partial class Reflex : UserControl
    {
        public string Description;
        public Reflexive Reflexive;
        Map map;
        XmlReader xr;
        public int Offset;
        bool Loaded = true;
        bool Start = true;
        int width;

        public Reflex(Reflexive r, Map map, string Description, XmlReader xr, int Width)
        {
            InitializeComponent();
            this.Reflexive = r;
            this.map = map;
            this.Description = Description;
            this.xr = xr;
            this.Offset = r.Offset;
            this.width = Width;
            this.Controls[0].Width = width;
            this.comboBox1.Width = width - 6;
            LoadControls();
        }

        public void LoadControls()
        {
            #region Xml Reading
            xr.ReadStartElement();
            panelEx1.Hide();
            panelEx1.Refresh();
            panelEx1.SuspendLayout();
            List<Control> Controls = new List<Control>();
            while (xr.Read())
            {
                switch (xr.NodeType)
                {
                    case XmlNodeType.Element:
                        {
                            switch (xr.Name.ToLower())
                            {
                                case "byte":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        Value b;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            b = new Value(Description, map.br.ReadByte(), typeof(byte), Offset);
                                        }
                                        else
                                        {
                                            b = new Value(Description, 0, typeof(byte), Offset);
                                        }
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "short":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        Value b;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            b = new Value(Description, map.br.ReadInt16(), typeof(short), Offset);
                                        }
                                        else
                                        {
                                            b = new Value(Description, 0, typeof(short), Offset);
                                        }
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "ushort":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        Value b;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            b = new Value(Description, map.br.ReadUInt16(), typeof(ushort), Offset);
                                        }
                                        else
                                        {
                                            b = new Value(Description, 0, typeof(ushort), Offset);
                                        }
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "int":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        Value b;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            b = new Value(Description, map.br.ReadInt32(), typeof(int), Offset);
                                        }
                                        else
                                        {
                                            b = new Value(Description, 0, typeof(int), Offset);
                                        }
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "uint":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        Value b;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            b = new Value(Description, map.br.ReadUInt32(), typeof(uint), Offset);
                                        }
                                        else
                                        {
                                            b = new Value(Description, 0, typeof(uint), Offset);
                                        }
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "float":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        Value b;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            b = new Value(Description, map.br.ReadSingle(), typeof(float), Offset);
                                        }
                                        else
                                        {
                                            b = new Value(Description, 0, typeof(float), Offset);
                                        }
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "string":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        int Length = Convert.ToInt32(xr.GetAttribute("length"));
                                        Value b;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            b = new Value(Description, map.br.ReadChars(Length), typeof(char[]), Offset);
                                        }
                                        else
                                        {
                                            b = new Value(Description, new char[Length], typeof(char[]), Offset);
                                        }
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "unknown":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        Value b;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            b = new Value(Description, map.br.ReadInt32(), typeof(int), Offset);
                                        }
                                        else
                                        {
                                            b = new Value(Description, 0, typeof(int), Offset);
                                        }
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "ident":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + Reflexive.Translation;
                                        TagReference t = new TagReference();
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset;
                                            t.Read(map, Offset, true);
                                        }
                                        else
                                        {
                                            t.Tag = new Tag();
                                            t.Class = "";
                                            t.Tag.Path = "";
                                            t.Tag.Identifier = -1;
                                            t.Offset = Offset;
                                        }
                                        Ident i = new Ident(t, map, Description);
                                        Controls.Add(i);
                                        i.Dock = DockStyle.Top;
                                        i.BringToFront();
                                        break;
                                    }
                                case "reflexive":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + Reflexive.Translation;
                                        int Size = Convert.ToInt32(xr.GetAttribute("size"));
                                        Reflexive r = new Reflexive();
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset;
                                            r.ChunkSize = Size;
                                            r.Read(map, Offset, true);
                                        }
                                        else
                                        {
                                            r.ChunkCount = -1;
                                            r.Translation = 0;
                                            r.Offset = Offset;
                                        }
                                        Reflex rr = new Reflex(r, map, Description, xr.ReadSubtree(), this.Controls[0].Width - 20);
                                        Controls.Add(rr);
                                        rr.Dock = DockStyle.Top;
                                        rr.BringToFront();
                                        break;
                                    }
                                case "stringid":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset")) + Reflexive.Translation;
                                        ScriptReference s = new ScriptReference();
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset;
                                            s.Read(map, Offset, true);
                                        }
                                        else
                                        {
                                            s.Reference = "";
                                            s.Offset = Offset;
                                        }
                                        StringID sid = new StringID(s, map, Description);
                                        Controls.Add(sid);
                                        sid.Dock = DockStyle.Top;
                                        sid.BringToFront();
                                        break;
                                    }
                                case "enum8":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        int Index = -1;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            Index = map.br.ReadByte();
                                        }
                                        List<string> Items = new List<string>();
                                        List<int> Values = new List<int>();
                                        XmlReader x = xr.ReadSubtree();
                                        while (x.Read())
                                        {
                                            switch (x.Name.ToLower())
                                            {
                                                case "option":
                                                    {
                                                        Items.Add(xr.GetAttribute("name"));
                                                        Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                        break;
                                                    }
                                            }
                                        }
                                        Enum e = new Enum(typeof(byte), Offset, Description, Items, Values, Index);
                                        Controls.Add(e);
                                        e.Dock = DockStyle.Top;
                                        e.BringToFront();
                                        break;
                                    }
                                case "enum16":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        int Index = -1;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            Index = map.br.ReadInt16();
                                        }
                                        List<string> Items = new List<string>();
                                        List<int> Values = new List<int>();
                                        XmlReader x = xr.ReadSubtree();
                                        while (x.Read())
                                        {
                                            switch (x.Name.ToLower())
                                            {
                                                case "option":
                                                    {
                                                        Items.Add(xr.GetAttribute("name"));
                                                        Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                        break;
                                                    }
                                            }
                                        }
                                        Enum e = new Enum(typeof(short), Offset, Description, Items, Values, Index);
                                        Controls.Add(e);
                                        e.Dock = DockStyle.Top;
                                        e.BringToFront();
                                        break;
                                    }
                                case "enum32":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        int Index = -1;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            Index = map.br.ReadInt32();
                                        }
                                        List<string> Items = new List<string>();
                                        List<int> Values = new List<int>();
                                        XmlReader x = xr.ReadSubtree();
                                        while (x.Read())
                                        {
                                            switch (x.Name.ToLower())
                                            {
                                                case "option":
                                                    {
                                                        Items.Add(xr.GetAttribute("name"));
                                                        Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                        break;
                                                    }
                                            }
                                        }
                                        Enum e = new Enum(typeof(int), Offset, Description, Items, Values, Index);
                                        Controls.Add(e);
                                        e.Dock = DockStyle.Top;
                                        e.BringToFront();
                                        break;
                                    }
                                case "bitmask8":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        int Index = -1;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            Index = map.br.ReadByte();
                                        }
                                        List<string> Items = new List<string>();
                                        List<int> Values = new List<int>();
                                        XmlReader x = xr.ReadSubtree();
                                        while (x.Read())
                                        {
                                            switch (x.Name.ToLower())
                                            {
                                                case "option":
                                                    {
                                                        Items.Add(xr.GetAttribute("name"));
                                                        Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                        break;
                                                    }
                                            }
                                        }
                                        Bitmask b = new Bitmask(Description, Index, Items, Values, Offset, typeof(byte));
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "bitmask16":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        ushort Index = 0;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            Index = map.br.ReadUInt16();
                                        }
                                        List<string> Items = new List<string>();
                                        List<int> Values = new List<int>();
                                        XmlReader x = xr.ReadSubtree();
                                        while (x.Read())
                                        {
                                            switch (x.Name.ToLower())
                                            {
                                                case "option":
                                                    {
                                                        Items.Add(xr.GetAttribute("name"));
                                                        Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                        break;
                                                    }
                                            }
                                        }
                                        Bitmask b = new Bitmask(Description, Index, Items, Values, Offset, typeof(short));
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "bitmask32":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        uint Index = 0;
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            Index = map.br.ReadUInt32();
                                        }
                                        List<string> Items = new List<string>();
                                        List<int> Values = new List<int>();
                                        XmlReader x = xr.ReadSubtree();
                                        while (x.Read())
                                        {
                                            switch (x.Name.ToLower())
                                            {
                                                case "option":
                                                    {
                                                        Items.Add(xr.GetAttribute("name"));
                                                        Values.Add(Convert.ToInt32(xr.GetAttribute("value")));
                                                        break;
                                                    }
                                            }
                                        }
                                        Bitmask b = new Bitmask(Description, Index, Items, Values, Offset, typeof(int));
                                        Controls.Add(b);
                                        b.Dock = DockStyle.Top;
                                        b.BringToFront();
                                        break;
                                    }
                                case "string32":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        string Value = "";
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            Value = new string(map.br.ReadChars(32));
                                        }
                                        String s = new String(Description, Offset, Value, StringType.String32);
                                        Controls.Add(s);
                                        s.Dock = DockStyle.Top;
                                        s.BringToFront();
                                        break;
                                    }
                                case "string64":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        string Value = "";
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            Value = new string(map.br.ReadChars(64));
                                        }
                                        String s = new String(Description, Offset, Value, StringType.String64);
                                        Controls.Add(s);
                                        s.Dock = DockStyle.Top;
                                        s.BringToFront(); 
                                        break;
                                    }
                                case "unicode64":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        string Value = "";
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            for (int i = 0; i < 32; i++)
                                            {
                                                Value += Convert.ToChar(map.br.ReadByte());
                                                map.br.BaseStream.Position++;
                                            }
                                        }
                                        String s = new String(Description, Offset, Value, StringType.Unicode64);
                                        Controls.Add(s);
                                        s.Dock = DockStyle.Top;
                                        s.BringToFront();
                                        break;
                                    }
                                case "unicode256":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        string Value = "";
                                        if (Reflexive.ChunkCount > 0)
                                        {
                                            map.br.BaseStream.Position = Offset + Reflexive.Translation;
                                            for (int i = 0; i < 128; i++)
                                            {
                                                Value += Convert.ToChar(map.br.ReadByte());
                                                map.br.BaseStream.Position++;
                                            }
                                        }
                                        String s = new String(Description, Offset, Value, StringType.Unicode256);
                                        Controls.Add(s);
                                        s.Dock = DockStyle.Top;
                                        s.BringToFront();
                                        break;
                                    }
                            }
                            break;
                        }
                }
            }

            //Load Controls
            for (int i = Controls.Count - 1; i > -1; i--)
            {
                panelEx1.Controls.Add(Controls[i]);
            }
            panelEx1.ResumeLayout(true);
            panelEx1.Show();
            #endregion
        }

        private void Reflex_Load(object sender, EventArgs e)
        {
            expandablePanel1.TitleText = Description;
            if (Reflexive.ChunkCount != 0)
            {
                comboBox1.Items.Add("0");

                //Select Index
                comboBox1.SelectedIndex = 0;
                Loaded = false;
                Start = false;

                //Expand
                expandablePanel1.Expanded = true;
            }
            else
            {
                //Expand
                expandablePanel1.Expanded = false;
            }
            this.Controls[0].Width = width;
            this.comboBox1.Width = width - 6;
        }

        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (Start == true)
            {
                return;
            }
            if (Loaded == false)
            {
                for (int i = 1; i < Reflexive.ChunkCount; i++)
                {
                    comboBox1.Items.Add(i.ToString());
                }
            }
            Loaded = true;

            #region Index Changing
            int Index = comboBox1.SelectedIndex;
            int Size = Reflexive.ChunkSize;
            for (int i = 0; i < panelEx1.Controls.Count; i++)
            {
                if (panelEx1.Controls[i].GetType() == typeof(Value))
                {
                    Value v = (Value)panelEx1.Controls[i];
                    map.br.BaseStream.Position = v.Offset + Reflexive.Translation + (Index * Size);
                    if (v.t == typeof(byte))
                    {
                        v.Object = map.br.ReadByte();
                    }
                    else if (v.t == typeof(short))
                    {
                        v.Object = map.br.ReadInt16();
                    }
                    else if (v.t == typeof(ushort))
                    {
                        v.Object = map.br.ReadUInt16();
                    }
                    else if (v.t == typeof(int))
                    {
                        v.Object = map.br.ReadInt32();
                    }
                    else if (v.t == typeof(uint))
                    {
                        v.Object = map.br.ReadUInt32();
                    }
                    else if (v.t == typeof(float))
                    {
                        v.Object = map.br.ReadSingle();
                    }
                    else if (v.t == typeof(char[]))
                    {
                        v.Object = map.br.ReadChars(v.Length);
                    }
                    v.Update();
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Ident))
                {
                    Ident id = (Ident)panelEx1.Controls[i];
                    map.br.BaseStream.Position = id.Offset + (Index * Size);
                    TagReference t = new TagReference();
                    t.Read(map, true);
                    id.Tag = t;
                    id.Update();
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Reflex))
                {
                    Reflex r = (Reflex)panelEx1.Controls[i];
                    int Offset = r.Offset + (Index * Size);
                    r.Update(Offset);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(StringID))
                {
                    StringID sid = (StringID)panelEx1.Controls[i];
                    ScriptReference s = new ScriptReference();
                    s.Read(map, sid.Offset + (Index * Size), true);
                    sid.Script = s;
                    sid.Update();
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Enum))
                {
                    Enum en = (Enum)panelEx1.Controls[i];
                    map.br.BaseStream.Position = en.Offset + Reflexive.Translation + (Index * Size);
                    object Value = 0;
                    if (en.t == typeof(byte))
                    {
                        Value = map.br.ReadByte();
                    }
                    else if (en.t == typeof(short))
                    {
                        Value = map.br.ReadInt16();
                    }
                    else if (en.t == typeof(int))
                    {
                        Value = map.br.ReadInt32();
                    }
                    en.Update(Value);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Bitmask))
                {
                    Bitmask b = (Bitmask)panelEx1.Controls[i];
                    map.br.BaseStream.Position = b.Offset + Reflexive.Translation + (Index * Size);
                    object Value = 0;
                    if (b.t == typeof(byte))
                    {
                        Value = map.br.ReadByte();
                    }
                    else if (b.t == typeof(short))
                    {
                        Value = map.br.ReadUInt16();
                    }
                    else if (b.t == typeof(int))
                    {
                        Value = map.br.ReadUInt32();
                    }
                    b.Update(Value);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(String))
                {
                    String s = (String)panelEx1.Controls[i];
                    map.br.BaseStream.Position = s.Offset + Reflexive.Translation + (Index * Size);
                    string Val = "";
                    if (s.t == StringType.String32)
                    {
                        Val = new string(map.br.ReadChars(32));
                    }
                    else if (s.t == StringType.String64)
                    {
                        Val = new string(map.br.ReadChars(64));
                    }
                    else if (s.t == StringType.Unicode64)
                    {
                        for (int x = 0; x < 32; x++)
                        {
                            Val += map.br.ReadChar();
                            map.br.BaseStream.Position++;
                        }
                    }
                    else if (s.t == StringType.Unicode256)
                    {
                        for (int x = 0; x < 128; x++)
                        {
                            Val += map.br.ReadChar();
                            map.br.BaseStream.Position++;
                        }
                    }
                    s.Update(Val);
                }
            }
            #endregion
        }

        public void Update(int Offset)
        {
            #region Update
            int T = Reflexive.Translation;
            Reflexive.Read(map, Offset, true);
            if (Reflexive.Translation < 0)
            {
                this.expandablePanel1.Expanded = false;
                return;
            }
            if (T < 0)
            {
                T = 0;
            }
            for (int i = 0; i < panelEx1.Controls.Count; i++)
            {
                if (panelEx1.Controls[i].GetType() == typeof(Value))
                {
                    Value v = (Value)panelEx1.Controls[i];
                    map.br.BaseStream.Position = Reflexive.Translation + v.Offset;
                    if (v.t == typeof(byte))
                    {
                        v.Object = map.br.ReadByte();
                    }
                    else if (v.t == typeof(short))
                    {
                        v.Object = map.br.ReadInt16();
                    }
                    else if (v.t == typeof(ushort))
                    {
                        v.Object = map.br.ReadUInt16();
                    }
                    else if (v.t == typeof(int))
                    {
                        v.Object = map.br.ReadInt32();
                    }
                    else if (v.t == typeof(uint))
                    {
                        v.Object = map.br.ReadUInt32();
                    }
                    else if (v.t == typeof(float))
                    {
                        v.Object = map.br.ReadSingle();
                    }
                    else if (v.t == typeof(char[]))
                    {
                        v.Object = map.br.ReadChars(v.Length);
                    }
                    v.Update();
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Ident))
                {
                    Ident id = (Ident)panelEx1.Controls[i];
                    map.br.BaseStream.Position = Reflexive.Translation + ((id.Offset - T));
                    TagReference t = new TagReference();
                    t.Read(map, true);
                    id.Tag.Tag = t.Tag;
                    id.Update();
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Reflex))
                {
                    Reflex r = (Reflex)panelEx1.Controls[i];
                    r.Update(Reflexive.Translation + ((r.Offset - T)));
                }
                else if (panelEx1.Controls[i].GetType() == typeof(StringID))
                {
                    StringID sid = (StringID)panelEx1.Controls[i];
                    ScriptReference s = new ScriptReference();
                    s.Read(map, Reflexive.Translation + ((sid.Offset - T)), true);
                    sid.Script = s;
                    sid.Update();
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Enum))
                {
                    Enum en = (Enum)panelEx1.Controls[i];
                    map.br.BaseStream.Position = Reflexive.Translation + en.Offset;
                    object Value = 0;
                    if (en.t == typeof(byte))
                    {
                        Value = map.br.ReadByte();
                    }
                    else if (en.t == typeof(short))
                    {
                        Value = map.br.ReadInt16();
                    }
                    else if (en.t == typeof(int))
                    {
                        Value = map.br.ReadInt32();
                    }
                    en.Update(Value);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Bitmask))
                {
                    Bitmask b = (Bitmask)panelEx1.Controls[i];
                    map.br.BaseStream.Position = Reflexive.Translation + b.Offset;
                    object Value = 0;
                    if (b.t == typeof(byte))
                    {
                        Value = map.br.ReadByte();
                    }
                    else if (b.t == typeof(short))
                    {
                        Value = map.br.ReadUInt16();
                    }
                    else if (b.t == typeof(int))
                    {
                        Value = map.br.ReadUInt32();
                    }
                    b.Update(Value);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(String))
                {
                    String s = (String)panelEx1.Controls[i];
                    map.br.BaseStream.Position = Reflexive.Translation + s.Offset;
                    string Val = "";
                    if (s.t == StringType.String32)
                    {
                        Val = new string(map.br.ReadChars(32));
                    }
                    else if (s.t == StringType.String64)
                    {
                        Val = new string(map.br.ReadChars(64));
                    }
                    else if (s.t == StringType.Unicode64)
                    {
                        for (int x = 0; x < 32; x++)
                        {
                            Val += map.br.ReadChar();
                            map.br.BaseStream.Position++;
                        }
                    }
                    else if (s.t == StringType.Unicode256)
                    {
                        for (int x = 0; x < 128; x++)
                        {
                            Val += map.br.ReadChar();
                            map.br.BaseStream.Position++;
                        }
                    }
                    s.Update(Val);
                }
            }
            #endregion
        }

        public void Save()
        {
            #region Saving
            if (comboBox1.SelectedIndex == -1)
                return;
            int Index = comboBox1.SelectedIndex;
            int Size = Reflexive.ChunkSize;
            for (int i = 0; i < panelEx1.Controls.Count; i++)
            {
                if (panelEx1.Controls[i].GetType() == typeof(Value))
                {
                    Value v = (Value)panelEx1.Controls[i];
                    map.bw.BaseStream.Position = v.Offset + Reflexive.Translation + (Index * Size);
                    object Val = v.Save();
                    if (v.t == typeof(byte))
                    {
                        map.bw.Write(byte.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(short))
                    {
                        map.bw.Write(short.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(ushort))
                    {
                        map.bw.Write(ushort.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(int))
                    {
                        map.bw.Write(int.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(uint))
                    {
                        map.bw.Write(uint.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(float))
                    {
                        map.bw.Write(float.Parse(Val.ToString()));
                    }
                    else if (v.t == typeof(char[]))
                    {
                        string s = Val.ToString();
                        map.bw.Write(s.ToCharArray());
                    }
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Ident))
                {
                    Ident id = (Ident)panelEx1.Controls[i];
                    TagReference t = id.Save();
                    t.Write(map, t.Offset + (Index * Size), true);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Reflex))
                {
                    Reflex r = (Reflex)panelEx1.Controls[i];
                    r.Save();
                }
                else if (panelEx1.Controls[i].GetType() == typeof(StringID))
                {
                    StringID sid = (StringID)panelEx1.Controls[i];
                    ScriptReference s = sid.Save();
                    s.Write(map, true);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Enum))
                {
                    Enum e = (Enum)panelEx1.Controls[i];
                    map.bw.BaseStream.Position = e.Offset + Reflexive.Translation + (Index * Size);
                    object Val = e.Save();
                    if (e.t == typeof(byte))
                    {
                        try { map.bw.Write(byte.Parse(Val.ToString())); }
                        catch (Exception) { map.bw.Write((byte)255); }
                    }
                    else if (e.t == typeof(short))
                    {
                        map.bw.Write(short.Parse(Val.ToString()));
                    }
                    else if (e.t == typeof(int))
                    {
                        map.bw.Write(int.Parse(Val.ToString()));
                    }
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Bitmask))
                {
                    Bitmask b = (Bitmask)panelEx1.Controls[i];
                    map.bw.BaseStream.Position = b.Offset + Reflexive.Translation + (Index * Size);
                    uint Value = b.Save();
                    object Val = Value;
                    if (b.t == typeof(byte))
                    {
                        map.bw.Write(byte.Parse(Val.ToString()));
                    }
                    else if (b.t == typeof(short))
                    {
                        map.bw.Write(ushort.Parse(Val.ToString()));
                    }
                    else if (b.t == typeof(int))
                    {
                        map.bw.Write(uint.Parse(Val.ToString()));
                    }
                }
                else if (panelEx1.Controls[i].GetType() == typeof(String))
                {
                    String s = (String)panelEx1.Controls[i];
                    map.bw.BaseStream.Position = s.Offset + Reflexive.Translation + (Index * Size);
                    if (s.t == StringType.String32)
                    {
                        map.bw.Write(s.Save().ToCharArray());
                        for (int x = s.Value.Length; x < 32; x++)
                        {
                            map.bw.Write((byte)0);
                        }
                    }
                    else if (s.t == StringType.String64)
                    {
                        map.bw.Write(s.Save().ToCharArray());
                        for (int x = s.Value.Length; x < 64; x++)
                        {
                            map.bw.Write((byte)0);
                        }
                    }
                    else if (s.t == StringType.Unicode64)
                    {
                        for (int x = 0; x < 32; x++)
                        {
                            try { map.bw.Write(s.Value[x]); }
                            catch { map.bw.Write((byte)0); }
                            map.bw.Write((byte)0);
                        }
                    }
                    else if (s.t == StringType.Unicode256)
                    {
                        for (int x = 0; x < 128; x++)
                        {
                            try { map.bw.Write(s.Value[x]); }
                            catch { map.bw.Write((byte)0); }
                            map.bw.Write((byte)0);
                        }
                    }
                }
            }
            #endregion
        }

        private void comboBox1_DropDown(object sender, EventArgs e)
        {
            if (Loaded == false)
            {
                for (int i = 1; i < Reflexive.ChunkCount; i++)
                {
                    comboBox1.Items.Add(i.ToString());
                }
            }
            Loaded = true;
        }

        public void Poke(ProcessStream p)
        {
            for (int i = 0; i < panelEx1.Controls.Count; i++)
            {
                if (panelEx1.Controls[i].GetType() == typeof(Value))
                {
                    Value v = (Value)panelEx1.Controls[i];
                    v.Poke(p, Reflexive.Translation);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Ident))
                {
                    Ident id = (Ident)panelEx1.Controls[i];
                    id.Poke(p, Reflexive.Translation);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Reflex))
                {
                    Reflex r = (Reflex)panelEx1.Controls[i];
                    r.Poke(p);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Enum))
                {
                    Enum e = (Enum)panelEx1.Controls[i];
                    e.Poke(p, Reflexive.Translation);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(Bitmask))
                {
                    Bitmask b = (Bitmask)panelEx1.Controls[i];
                    b.Poke(p, Reflexive.Translation);
                }
                else if (panelEx1.Controls[i].GetType() == typeof(String))
                {
                    String s = (String)panelEx1.Controls[i];
                    s.Poke(p, Reflexive.Translation);
                }
            }
        }
    }
}
