﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing.Imaging;
using System.Drawing;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.IO;
using ComponentAce.Compression.Libs.zlib;

namespace HaloMap
{
    public class Bitmap
    {
        public static ColorRGBA[] H2Palette = LoadHalo2Palette();

        public struct ColorRGBA
        {
            public byte r;
            public byte g;
            public byte b;
            public byte a;
        }

        public static ColorRGBA intToRGBA(uint color)
        {
            ColorRGBA rc;
            rc.r = (byte)((((color >> 11) & 31) * 255) / 31);
            rc.g = (byte)((((color >> 5) & 63) * 255) / 63);
            rc.b = (byte)((((color >> 0) & 31) * 255) / 31);
            rc.a = 255;
            return rc;
        }

        public static int RGBAToint(ColorRGBA c)
        {
            return (c.a << 24) | (c.r << 16) | (c.g << 8) | c.b;
        }

        public static ColorRGBA GradientColors(ColorRGBA Col1, ColorRGBA Col2)
        {
            ColorRGBA ret;
            ret.r = (byte)(((Col1.r * 2 + Col2.r)) / 3);
            ret.g = (byte)(((Col1.g * 2 + Col2.g)) / 3);
            ret.b = (byte)(((Col1.b * 2 + Col2.b)) / 3);
            ret.a = 255;
            return ret;
        }

        public static ColorRGBA GradientColorsHalf(ColorRGBA Col1, ColorRGBA Col2)
        {
            ColorRGBA ret;
            ret.r = (byte)(Col1.r / 2 + Col2.r / 2);
            ret.g = (byte)(Col1.g / 2 + Col2.g / 2);
            ret.b = (byte)(Col1.b / 2 + Col2.b / 2);
            ret.a = 255;
            return ret;
        }

        public static System.Drawing.Bitmap LoadPreview(Map map)
        {
            //Read the Bitmap Raw
            BitmapRaw.Read(map);

            //Load Preview
            return LoadPreview(map, BitmapRaw.RawChunks[0]);
        }
        public static System.Drawing.Bitmap LoadPreview(Map map, BitmapChunk Bitmap)
        {
            #region Preview
            byte[] DecodedLOD = Decompress(Bitmap.SourceData);
            byte[] RecodedLOD = new byte[0];
            byte[] DecompressedData = Decompress(Bitmap.SourceData);
            PixelFormat pixelFormat = PixelFormat.Undefined;
            int LengthMultipier = 0;
            int HalfSourceDataLength;
            //if (Bitmap.Width % 8 != 0)
            //{
            //    Bitmap.Width = (ushort)(Bitmap.Width + (8 - (Bitmap.Width % 8)));
            //}
            switch (Bitmap.BitmapFormat)
            {
                case BitmapFormat.DXT1:
                    {
                        DecodedLOD = DecodeDXT1(Bitmap.Height, Bitmap.Width, DecompressedData);
                        pixelFormat = PixelFormat.Format32bppArgb;
                        LengthMultipier = 4;
                        break;
                    }
                case BitmapFormat.DXT3:
                    {
                        DecodedLOD = DecodeDXT23(Bitmap.Height, Bitmap.Width, DecompressedData);
                        pixelFormat = PixelFormat.Format32bppArgb;
                        LengthMultipier = 4;
                        break;
                    }
                case BitmapFormat.DXT5:
                    {
                        DecodedLOD = DecodeDXT45(Bitmap.Height, Bitmap.Width, DecompressedData);
                        pixelFormat = PixelFormat.Format32bppArgb;
                        LengthMultipier = 4;
                        break;
                    }
                case BitmapFormat.A8R8G8B8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 32, true);
                    }
                    pixelFormat = PixelFormat.Format32bppArgb;
                    LengthMultipier = 4;
                    break;
                case BitmapFormat.X8R8G8B8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.Width, Bitmap.Height, -1, 32, true);
                    }
                    pixelFormat = PixelFormat.Format32bppRgb;
                    LengthMultipier = 4;
                    break;
                case BitmapFormat.A4R4G4B4:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, -1, 16, true);
                    }
                    LengthMultipier = 2;
                    pixelFormat = PixelFormat.Format32bppArgb;
                    int QuaterSourceDataLength = DecodedLOD.Length / 4;
                    RecodedLOD = new byte[DecompressedData.Length * 2];
                    for (int i = 0; i < QuaterSourceDataLength; i++)
                    {
                        int a = i * 4;
                        RecodedLOD[a] = (byte)(DecodedLOD[a] & 0XF0);
                        RecodedLOD[a + 1] = (byte)(DecodedLOD[a] & 0X0F);
                        RecodedLOD[a + 2] = (byte)(DecodedLOD[a + 1] & 0XF0);
                        RecodedLOD[a + 3] = (byte)(DecodedLOD[a + 1] & 0X0F);
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                case BitmapFormat.Unknown:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, 0, Bitmap.Width, Bitmap.Height, -1, 16, true);
                    }
                    LengthMultipier = 2;
                    pixelFormat = PixelFormat.Format32bppArgb;
                    HalfSourceDataLength = DecompressedData.Length / 2;
                    RecodedLOD = new byte[HalfSourceDataLength * 4];
                    for (int i = 0; i < HalfSourceDataLength; i++)
                    {
                        int a = i * 2;
                        RecodedLOD[a * 2] = DecodedLOD[a + 1];
                        RecodedLOD[(a * 2) + 1] = DecodedLOD[a + 1];
                        RecodedLOD[(a * 2) + 2] = DecodedLOD[a + 1];
                        RecodedLOD[(a * 2) + 3] = DecodedLOD[a];
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                case BitmapFormat.A1R5G5B5:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, -1, 16, true);
                    }
                    LengthMultipier = 2;
                    pixelFormat = PixelFormat.Format16bppArgb1555;
                    break;
                case BitmapFormat.R5G6B5:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, -1, 16, true);
                    }
                    LengthMultipier = 2;
                    pixelFormat = PixelFormat.Format16bppRgb565;
                    break;
                case BitmapFormat.A8Y8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 16, true);
                    }
                    LengthMultipier = 4;
                    pixelFormat = PixelFormat.Format32bppArgb;
                    HalfSourceDataLength = DecompressedData.Length / 2;
                    RecodedLOD = new byte[HalfSourceDataLength * 4];
                    for (int i = 0; i < HalfSourceDataLength; i++)
                    {
                        int a = i * 2;
                        RecodedLOD[a] = DecodedLOD[a];
                        RecodedLOD[a + 1] = DecodedLOD[a];
                        RecodedLOD[a + 2] = DecodedLOD[a];
                        RecodedLOD[a + 3] = DecodedLOD[a + 1];
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                case BitmapFormat.P8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 8, true);
                    }
                    LengthMultipier = 4;
                    pixelFormat = PixelFormat.Format32bppArgb;

                    RecodedLOD = new byte[DecompressedData.Length * 4];
                    for (int i = 0; i < DecompressedData.Length; i++)
                    {
                        int a = i * 4;
                        RecodedLOD[a] = H2Palette[DecodedLOD[i]].r;
                        RecodedLOD[a + 1] = H2Palette[DecodedLOD[i]].g;
                        RecodedLOD[a + 2] = H2Palette[DecodedLOD[i]].b;
                        RecodedLOD[a + 3] = H2Palette[DecodedLOD[i]].a;
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                case BitmapFormat.Y8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 8, true);
                    }
                    LengthMultipier = 4;
                    pixelFormat = PixelFormat.Format32bppArgb;
                    RecodedLOD = new byte[DecompressedData.Length * 4];
                    for (int i = 0; i < DecompressedData.Length; i++)
                    {
                        int a = i * 4;
                        RecodedLOD[a] = DecodedLOD[i];
                        RecodedLOD[a + 1] = DecodedLOD[i];
                        RecodedLOD[a + 2] = DecodedLOD[i];
                        RecodedLOD[a + 3] = 255;
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                //case BitmapFormat.Lightmap://Working
                //    if (Bitmap.Swizzle)
                //    {
                //        DecodedLOD = Bitmaps.Swizzle(DecodedLOD, PixelOffset, width, height, Depth, 8, true);
                //    }
                //    LengthMultipier = 4;
                //    pixelFormat = PixelFormat.Format32bppArgb;

                //    RecodedLOD = new byte[DecodedLOD.Length * 4];
                //    LightMap.Palette Palette = Map.Sbsp[SBSPIndex, Tags.SearchType.Index].LightMap.Palettes[PaletteIndex];
                //    for (int i = 0; i < DecodedLOD.Length; i++)
                //    {
                //        int a = i * 4;
                //        RecodedLOD[a + 0] = Palette.Colors[DecodedLOD[i]].R;
                //        RecodedLOD[a + 1] = Palette.Colors[DecodedLOD[i]].G;
                //        RecodedLOD[a + 2] = Palette.Colors[DecodedLOD[i]].B;
                //        RecodedLOD[a + 3] = Palette.Colors[DecodedLOD[i]].A;
                //    }
                //    DecodedLOD = RecodedLOD;
                //    break;
                case BitmapFormat.AY8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 8, true);
                    }
                    LengthMultipier = 4;
                    pixelFormat = PixelFormat.Format32bppArgb;

                    RecodedLOD = new byte[DecodedLOD.Length * 4];
                    for (int i = 0; i < DecodedLOD.Length; i++)
                    {
                        int a = i * 4;
                        RecodedLOD[a] = DecodedLOD[i];
                        RecodedLOD[a + 1] = DecodedLOD[i];
                        RecodedLOD[a + 2] = DecodedLOD[i];
                        RecodedLOD[a + 3] = 255;
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                case BitmapFormat.A8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 8, true);
                    }
                    LengthMultipier = 4;
                    pixelFormat = PixelFormat.Format32bppArgb;
                    RecodedLOD = new byte[DecompressedData.Length * 4];
                    for (int i = 0; i < DecompressedData.Length; i++)
                    {
                        int a = i * 4;
                        RecodedLOD[a] = 255;
                        RecodedLOD[a + 1] = 255;
                        RecodedLOD[a + 2] = 255;
                        RecodedLOD[a + 3] = DecodedLOD[i];
                    }
                    DecodedLOD = RecodedLOD;
                    break;
            }
            System.Drawing.Bitmap bitmap = new System.Drawing.Bitmap(Bitmap.Width, Bitmap.Height);
            BitmapData data = bitmap.LockBits(new Rectangle(0, 0, Bitmap.Width, Bitmap.Height), ImageLockMode.ReadWrite, pixelFormat);
            Marshal.Copy(DecodedLOD, 0, data.Scan0, Bitmap.Width * Bitmap.Height * LengthMultipier);
            bitmap.UnlockBits(data);
            return bitmap;
            #endregion
        }

        public static byte[] Decode(Map map, BitmapChunk Bitmap)
        {
            #region Decode
            byte[] DecodedLOD = Decompress(Bitmap.SourceData);
            byte[] RecodedLOD = new byte[0];
            byte[] DecompressedData = Decompress(Bitmap.SourceData);
            PixelFormat pixelFormat = PixelFormat.Undefined;
            int LengthMultipier = 0;
            int HalfSourceDataLength;
            if (Bitmap.Width % 8 != 0)
            {
                Bitmap.Width = (ushort)(Bitmap.Width + (8 - (Bitmap.Width % 8)));
            }
            switch (Bitmap.BitmapFormat)
            {
                case BitmapFormat.A8R8G8B8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 32, true);
                    }
                    pixelFormat = PixelFormat.Format32bppArgb;
                    LengthMultipier = 4;
                    break;
                case BitmapFormat.X8R8G8B8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.Width, Bitmap.Height, -1, 32, true);
                    }
                    pixelFormat = PixelFormat.Format32bppRgb;
                    LengthMultipier = 4;
                    break;
                case BitmapFormat.A4R4G4B4:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, -1, 16, true);
                    }
                    LengthMultipier = 2;
                    pixelFormat = PixelFormat.Format32bppArgb;
                    int QuaterSourceDataLength = DecodedLOD.Length / 4;
                    RecodedLOD = new byte[DecompressedData.Length * 2];
                    for (int i = 0; i < QuaterSourceDataLength; i++)
                    {
                        int a = i * 4;
                        RecodedLOD[a] = (byte)(DecodedLOD[a] & 0XF0);
                        RecodedLOD[a + 1] = (byte)(DecodedLOD[a] & 0X0F);
                        RecodedLOD[a + 2] = (byte)(DecodedLOD[a + 1] & 0XF0);
                        RecodedLOD[a + 3] = (byte)(DecodedLOD[a + 1] & 0X0F);
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                case BitmapFormat.Unknown:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, 0, Bitmap.Width, Bitmap.Height, -1, 16, true);
                    }
                    LengthMultipier = 2;
                    pixelFormat = PixelFormat.Format32bppArgb;
                    HalfSourceDataLength = DecompressedData.Length / 2;
                    RecodedLOD = new byte[HalfSourceDataLength * 4];
                    for (int i = 0; i < HalfSourceDataLength; i++)
                    {
                        int a = i * 2;
                        RecodedLOD[a * 2] = DecodedLOD[a + 1];
                        RecodedLOD[(a * 2) + 1] = DecodedLOD[a + 1];
                        RecodedLOD[(a * 2) + 2] = DecodedLOD[a + 1];
                        RecodedLOD[(a * 2) + 3] = DecodedLOD[a];
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                case BitmapFormat.A1R5G5B5:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, -1, 16, true);
                    }
                    LengthMultipier = 2;
                    pixelFormat = PixelFormat.Format16bppArgb1555;
                    break;
                case BitmapFormat.R5G6B5:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, -1, 16, true);
                    }
                    LengthMultipier = 2;
                    pixelFormat = PixelFormat.Format16bppRgb565;
                    break;
                case BitmapFormat.A8Y8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 16, true);
                    }
                    LengthMultipier = 4;
                    pixelFormat = PixelFormat.Format32bppArgb;
                    HalfSourceDataLength = DecompressedData.Length / 2;
                    RecodedLOD = new byte[HalfSourceDataLength * 4];
                    for (int i = 0; i < HalfSourceDataLength; i++)
                    {
                        int a = i * 2;
                        RecodedLOD[a] = DecodedLOD[a];
                        RecodedLOD[a + 1] = DecodedLOD[a];
                        RecodedLOD[a + 2] = DecodedLOD[a];
                        RecodedLOD[a + 3] = DecodedLOD[a + 1];
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                case BitmapFormat.P8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 8, true);
                    }
                    LengthMultipier = 4;
                    pixelFormat = PixelFormat.Format32bppArgb;

                    RecodedLOD = new byte[DecompressedData.Length * 4];
                    for (int i = 0; i < DecompressedData.Length; i++)
                    {
                        int a = i * 4;
                        RecodedLOD[a] = H2Palette[DecodedLOD[i]].r;
                        RecodedLOD[a + 1] = H2Palette[DecodedLOD[i]].g;
                        RecodedLOD[a + 2] = H2Palette[DecodedLOD[i]].b;
                        RecodedLOD[a + 3] = H2Palette[DecodedLOD[i]].a;
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                case BitmapFormat.Y8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 8, true);
                    }
                    LengthMultipier = 4;
                    pixelFormat = PixelFormat.Format32bppArgb;
                    RecodedLOD = new byte[DecompressedData.Length * 4];
                    for (int i = 0; i < DecompressedData.Length; i++)
                    {
                        int a = i * 4;
                        RecodedLOD[a] = DecodedLOD[i];
                        RecodedLOD[a + 1] = DecodedLOD[i];
                        RecodedLOD[a + 2] = DecodedLOD[i];
                        RecodedLOD[a + 3] = 255;
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                //case BitmapFormat.Lightmap://Working
                //    if (Swizzle)
                //        DecodedLOD = Bitmaps.Swizzle(DecodedLOD, PixelOffset, width, height, Depth, 8, true);
                //    LengthMultipier = 4;
                //    pixelFormat = PixelFormat.Format32bppArgb;

                //    RecodedLOD = new byte[DecodedLOD.Length * 4];
                //    LightMap.Palette Palette = Map.Sbsp[SBSPIndex, Tags.SearchType.Index].LightMap.Palettes[PaletteIndex];
                //    for (int i = 0; i < DecodedLOD.Length; i++)
                //    {
                //        int a = i * 4;
                //        RecodedLOD[a + 0] = Palette.Colors[DecodedLOD[i]].R;
                //        RecodedLOD[a + 1] = Palette.Colors[DecodedLOD[i]].G;
                //        RecodedLOD[a + 2] = Palette.Colors[DecodedLOD[i]].B;
                //        RecodedLOD[a + 3] = Palette.Colors[DecodedLOD[i]].A;
                //    }
                //    DecodedLOD = RecodedLOD;
                //    break;
                case BitmapFormat.AY8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 8, true);
                    }
                    LengthMultipier = 4;
                    pixelFormat = PixelFormat.Format32bppArgb;

                    RecodedLOD = new byte[DecodedLOD.Length * 4];
                    for (int i = 0; i < DecodedLOD.Length; i++)
                    {
                        int a = i * 4;
                        RecodedLOD[a] = DecodedLOD[i];
                        RecodedLOD[a + 1] = DecodedLOD[i];
                        RecodedLOD[a + 2] = DecodedLOD[i];
                        RecodedLOD[a + 3] = 255;
                    }
                    DecodedLOD = RecodedLOD;
                    break;
                case BitmapFormat.A8:
                    if (Bitmap.Swizzle == true)
                    {
                        DecodedLOD = Swizzle(DecompressedData, Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 8, true);
                    }
                    LengthMultipier = 4;
                    pixelFormat = PixelFormat.Format32bppArgb;
                    RecodedLOD = new byte[DecompressedData.Length * 4];
                    for (int i = 0; i < DecompressedData.Length; i++)
                    {
                        int a = i * 4;
                        RecodedLOD[a] = 255;
                        RecodedLOD[a + 1] = 255;
                        RecodedLOD[a + 2] = 255;
                        RecodedLOD[a + 3] = DecodedLOD[i];
                    }
                    DecodedLOD = RecodedLOD;
                    break;
            }
            return DecodedLOD;
            #endregion
        }

        public static byte[] Encode(Map map, BitmapChunk Bitmap)
        {
            #region Encode
            byte[] byteArray = Bitmap.SourceData;
            int tmpHalf;
            byte[] tmpByteArray;

            switch (Bitmap.BitmapFormat)
            {
                case BitmapFormat.A8R8G8B8:
                    if (Bitmap.Swizzle) byteArray = Swizzle(byteArray,
                        Bitmap.Width, Bitmap.Height, Bitmap.Depth, 32, false);
                    break;

                case BitmapFormat.X8R8G8B8:
                    if (Bitmap.Swizzle) byteArray = Swizzle(byteArray,
                        Bitmap.Width, Bitmap.Height, Bitmap.Depth, 32, false);
                    break;

                //16 bit
                case BitmapFormat.A4R4G4B4:
                    if (Bitmap.Swizzle) byteArray = Swizzle(byteArray,
                        Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, -1, 16, false);
                    break;

                case BitmapFormat.Unknown:
                    tmpHalf = byteArray.Length * 2;
                    tmpByteArray = new byte[tmpHalf / 4];
                    for (int x = 0; x < (byteArray.Length / 4); x++)
                    {
                        int r = x * 2;
                        tmpByteArray[r + 1] = byteArray[r * 2 + 0];
                        tmpByteArray[r + 1] = byteArray[r * 2 + 1];
                        tmpByteArray[r + 1] = byteArray[r * 2 + 2];
                        tmpByteArray[r + 0] = byteArray[r * 2 + 3];
                    }
                    byteArray = tmpByteArray;

                    if (Bitmap.Swizzle) byteArray = Swizzle(byteArray,
                        0, Bitmap.Width, Bitmap.Height, -1, 16, false);
                    break;

                case BitmapFormat.A1R5G5B5:
                    if (Bitmap.Swizzle) byteArray = Swizzle(byteArray,
                        Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, -1, 16, false);
                    break;

                case BitmapFormat.R5G6B5:
                    if (Bitmap.Swizzle) byteArray = Swizzle(byteArray,
                        Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, -1, 16, false);
                    break;

                case BitmapFormat.A8Y8:
                    tmpHalf = byteArray.Length * 2;
                    tmpByteArray = new byte[tmpHalf / 4];
                    for (int x = 0; x < (byteArray.Length / 4); x++)
                    {
                        int r = x * 2;
                        tmpByteArray[r + 1] = byteArray[r + 0];
                        tmpByteArray[r + 1] = byteArray[r + 1];
                        tmpByteArray[r + 1] = byteArray[r + 2];
                        tmpByteArray[r + 0] = byteArray[r + 3];
                    }
                    byteArray = tmpByteArray;

                    if (Bitmap.Swizzle) byteArray = Swizzle(byteArray,
                        Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 16, false);
                    break;

                case BitmapFormat.P8:
                    tmpByteArray = new byte[byteArray.Length / 4];

                    ColorRGBA baColor = new ColorRGBA();
                    for (int x = 0; x < tmpByteArray.Length; x++)
                    {
                        int r = x * 4;
                        baColor.r = byteArray[r];
                        baColor.g = byteArray[r + 1];
                        baColor.b = byteArray[r + 2];
                        baColor.a = byteArray[r + 3];

                        byte colorIndx = 0;
                        float storedDiff = 0;// 195075f;
                        ColorRGBA palColor = new ColorRGBA();
                        for (byte y = 0; y < 255; y++)
                        {
                            palColor.r = H2Palette[y].r;
                            palColor.g = H2Palette[y].g;
                            palColor.b = H2Palette[y].b;
                            palColor.a = H2Palette[y].a;

                            float diff =
                                (((baColor.r - palColor.r) * (baColor.r - palColor.r)) +
                                ((baColor.g - palColor.g) * (baColor.g - palColor.g))) +
                                ((baColor.b - palColor.b) * (baColor.b - palColor.b));

                            if (diff <= storedDiff)
                            {
                                colorIndx = y;
                                storedDiff = diff;
                            }
                        }
                        tmpByteArray[x] = colorIndx;
                    }
                    byteArray = tmpByteArray;

                    if (Bitmap.Swizzle) byteArray = Swizzle(byteArray,
                        Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 8, false);
                    break;

                case BitmapFormat.Y8:
                    tmpByteArray = new byte[byteArray.Length / 4];
                    for (int x = 0; x < tmpByteArray.Length; x++)
                    {
                        int r = x * 4;
                        tmpByteArray[x] = byteArray[r + 0];
                        tmpByteArray[x] = byteArray[r + 1];
                        tmpByteArray[x] = byteArray[r + 2];
                        tmpByteArray[255] = byteArray[r + 3];
                    }
                    byteArray = tmpByteArray;

                    if (Bitmap.Swizzle) byteArray = Swizzle(byteArray,
                        Bitmap.PixelOffset, Bitmap.Width, Bitmap.Height, Bitmap.Depth, 8, false);
                    break;
            }

            return Compress(byteArray);
            #endregion
        }

        public static byte[] Decompress(byte[] SourceData)
        {
            //We need a new Memory stream so we can use ZLib
            MemoryStream Out = new MemoryStream();
            ZOutputStream ZOut = new ZOutputStream(Out);
            MemoryStream In = new MemoryStream(SourceData);

            //Decompress
            CopyStream(In, ZOut);

            //Close
            byte[] Data = Out.ToArray();
            Out.Close();
            ZOut.Close();
            In.Close();

            //Return
            return Data;
        }

        public static byte[] Compress(byte[] SourceData)
        {
            //We need a new Memory stream so we can use ZLib
            MemoryStream Out = new MemoryStream();
            ZOutputStream ZOut = new ZOutputStream(Out, zlibConst.Z_DEFAULT_COMPRESSION);
            MemoryStream In = new MemoryStream(SourceData);

            //Compress
            CopyStream(In, ZOut);

            //Close
            byte[] Data = Out.ToArray(); //ERROR!!!!
            Out.Close();
            ZOut.Close();
            In.Close();

            //Return
            return Data;
        }

        public static void CopyStream(Stream input, Stream output)
        {
            byte[] buffer = new byte[2000];
            int len;
            while ((len = input.Read(buffer, 0, 2000)) > 0)
            {
                output.Write(buffer, 0, len);
            }
            output.Flush();
        }

        public static byte[] DecodeDXT1(int height, int width, byte[] sourceData)
        {
            #region DXT1
            byte[] destData = new byte[height * width * 4];
            ColorRGBA[] color = new ColorRGBA[4];
            int dptr = 0;
            ColorRGBA CColor;
            int CData;
            int ChunksPerHLine = width / 4;
            bool trans;
            ColorRGBA zeroColor = new ColorRGBA();
            uint c1;
            uint c2;
            if (ChunksPerHLine == 0)
            {
                ChunksPerHLine += 1;
            }
            for (int i = 0; i < (width * height); i += 16)
            {
                c1 = (Convert.ToUInt32(sourceData[dptr + 1]) << 8) | (sourceData[dptr]);
                c2 = (Convert.ToUInt32(sourceData[dptr + 3]) << 8) | (sourceData[dptr + 2]);
                if (c1 > c2)
                {
                    trans = false;
                }
                else
                {
                    trans = true;
                }
                color[0] = intToRGBA(c1);
                color[1] = intToRGBA(c2);
                if (!(trans))
                {
                    color[2] = GradientColors(color[0], color[1]);
                    color[3] = GradientColors(color[1], color[0]);
                }
                else
                {
                    color[2] = GradientColorsHalf(color[0], color[1]);
                    color[3] = zeroColor;
                }
                CData = (Convert.ToInt32(sourceData[dptr + 4]) << 0) | (Convert.ToInt32(sourceData[dptr + 5]) << 8) | (Convert.ToInt32(sourceData[dptr + 6]) << 16) | (Convert.ToInt32(sourceData[dptr + 7]) << 24);
                int ChunkNum = i / 16;
                long XPos = ChunkNum % ChunksPerHLine;
                long YPos = (ChunkNum - XPos) / ChunksPerHLine;
                long ttmp;
                int sizeh = height < 4 ? height : 4;
                int sizew = width < 4 ? width : 4;
                for (int x = 0; x <= sizeh - 1; x++)
                {
                    for (int y = 0; y <= sizew - 1; y++)
                    {
                        CColor = color[CData & 3];
                        CData >>= 2;
                        ttmp = ((YPos * 4 + x) * width + XPos * 4 + y) * 4;
                        destData[ttmp] = CColor.b;
                        destData[ttmp + 1] = CColor.g;
                        destData[ttmp + 2] = CColor.r;
                        destData[ttmp + 3] = CColor.a;
                    }
                }
                dptr += 8;
            }
            return destData;
            #endregion
        }

        public static byte[] DecodeDXT23(int height, int width, byte[] sourceData)
        {
            #region DXT2&3
            byte[] destData = new byte[height * width * 4];
            ColorRGBA[] color = new ColorRGBA[4];
            ColorRGBA CColor;
            int CData;
            int ChunksPerHLine = width / 4;

            if (ChunksPerHLine == 0)
            {
                ChunksPerHLine += 1;
            }

            for (int i = 0; i <= (width * height) - 1; i += 16)
            {
                color[0] = intToRGBA(Convert.ToUInt32(sourceData[i + 8]) | Convert.ToUInt32(sourceData[i + 9]) << 8);
                color[1] = intToRGBA(Convert.ToUInt32(sourceData[i + 10]) | Convert.ToUInt32(sourceData[i + 11]) << 8);
                color[2] = GradientColors(color[0], color[1]);
                color[3] = GradientColors(color[1], color[0]);
                CData = (Convert.ToInt32(sourceData[i + 12]) << 0) | (Convert.ToInt32(sourceData[i + 13]) << 8) | (Convert.ToInt32(sourceData[i + 14]) << 16) | (Convert.ToInt32(sourceData[i + 15]) << 24);
                int ChunkNum = i / 16;
                long XPos = ChunkNum % ChunksPerHLine;
                long YPos = (ChunkNum - XPos) / ChunksPerHLine;
                long ttmp;
                int alpha;
                int sizeh = height < 4 ? height : 4;
                int sizew = width < 4 ? width : 4;

                for (int x = 0; x <= sizeh - 1; x++)
                {
                    alpha = sourceData[i + (2 * x)] | Convert.ToInt32(sourceData[i + (2 * x) + 1]) << 8;
                    for (int y = 0; y <= sizew - 1; y++)
                    {
                        CColor = color[CData & 3];
                        CData >>= 2;
                        CColor.a = (byte)((alpha & 15) * 16);
                        alpha >>= 4;
                        ttmp = ((YPos * 4 + x) * width + XPos * 4 + y) * 4;
                        destData[ttmp] = CColor.b;
                        destData[ttmp + 1] = CColor.g;
                        destData[ttmp + 2] = CColor.r;
                        destData[ttmp + 3] = CColor.a;
                    }
                }
            }
            return destData;
            #endregion
        }

        public static byte[] DecodeDXT45(int height, int width, byte[] sourceData)
        {
            #region DXT4&5
            byte[] destData = new byte[height * width * 4];
            ColorRGBA[] color = new ColorRGBA[4];
            ColorRGBA CColor;
            int CData;
            int ChunksPerHLine = width / 4;
            if (ChunksPerHLine == 0)
            {
                ChunksPerHLine += 1;
            }
            for (int i = 0; i <= (width * height) - 1; i += 16)
            {
                color[0] = intToRGBA(Convert.ToUInt32(sourceData[i + 8]) | Convert.ToUInt32(sourceData[i + 9]) << 8);
                color[1] = intToRGBA(Convert.ToUInt32(sourceData[i + 10]) | Convert.ToUInt32(sourceData[i + 11]) << 8);
                color[2] = GradientColors(color[0], color[1]);
                color[3] = GradientColors(color[1], color[0]);
                CData = (Convert.ToInt32(sourceData[i + 12]) << 0) | (Convert.ToInt32(sourceData[i + 13]) << 8) | (Convert.ToInt32(sourceData[i + 14]) << 16) | (Convert.ToInt32(sourceData[i + 15]) << 24);
                byte[] alpha = new byte[8];
                alpha[0] = sourceData[i];
                alpha[1] = sourceData[i + 1];
                if ((alpha[0] > alpha[1]))
                {
                    alpha[2] = (byte)((6 * alpha[0] + 1 * alpha[1] + 3) / 7);
                    alpha[3] = (byte)((5 * alpha[0] + 2 * alpha[1] + 3) / 7);
                    alpha[4] = (byte)((4 * alpha[0] + 3 * alpha[1] + 3) / 7);
                    alpha[5] = (byte)((3 * alpha[0] + 4 * alpha[1] + 3) / 7);
                    alpha[6] = (byte)((2 * alpha[0] + 5 * alpha[1] + 3) / 7);
                    alpha[7] = (byte)((1 * alpha[0] + 6 * alpha[1] + 3) / 7);
                }
                else
                {
                    alpha[2] = (byte)((4 * alpha[0] + 1 * alpha[1] + 2) / 5);
                    alpha[3] = (byte)((3 * alpha[0] + 2 * alpha[1] + 2) / 5);
                    alpha[4] = (byte)((2 * alpha[0] + 3 * alpha[1] + 2) / 5);
                    alpha[5] = (byte)((1 * alpha[0] + 4 * alpha[1] + 2) / 5);
                    alpha[6] = 0;
                    alpha[7] = 255;
                }
                long tmpdword;
                int tmpword;
                long alphaDat;
                tmpword = sourceData[i + 2] | (Convert.ToInt32(sourceData[i + 3]) << 8);
                tmpdword = sourceData[i + 4] | (Convert.ToInt32(sourceData[i + 5]) << 8) | (sourceData[i + 6] << 16) | (Convert.ToInt32(sourceData[i + 7]) << 24);
                alphaDat = tmpword | (tmpdword << 16);
                int ChunkNum = i / 16;
                long XPos = ChunkNum % ChunksPerHLine;
                long YPos = (ChunkNum - XPos) / ChunksPerHLine;
                long ttmp;
                int sizeh = height < 4 ? height : 4;
                int sizew = width < 4 ? width : 4;
                for (int x = 0; x <= sizeh - 1; x++)
                {
                    for (int y = 0; y <= sizew - 1; y++)
                    {
                        CColor = color[CData & 3];
                        CData >>= 2;
                        CColor.a = alpha[alphaDat & 7];
                        alphaDat >>= 3;
                        ttmp = ((YPos * 4 + x) * width + XPos * 4 + y) * 4;
                        destData[ttmp] = CColor.b;
                        destData[ttmp + 1] = CColor.g;
                        destData[ttmp + 2] = CColor.r;
                        destData[ttmp + 3] = CColor.a;
                    }
                }
            }
            return destData;
            #endregion
        }

        #region Swizzle
        public static byte[] Swizzle(byte[] raw, int pixOffset, int width, int height, int depth, int bitCount, bool deswizzle)
        {
            bitCount /= 8;
            int a = 0;
            int b = 0;
            byte[] dataArray = new byte[raw.Length]; //width * height * bitCount;

            MaskSet masks = new MaskSet(width, height, depth);
            pixOffset = 0;

            for (int y = 0; y < height; y++)
            {
                for (int x = 0; x < width; x++)
                {
                    if (deswizzle)
                    {
                        a = ((y * width) + x) * bitCount;
                        b = (Swizzle(x, y, -1, masks)) * bitCount;
                    }
                    else
                    {
                        b = ((y * width) + x) * bitCount;
                        a = (Swizzle(x, y, -1, masks)) * bitCount;
                    }

                    if (a < dataArray.Length && b < raw.Length)
                    {
                        for (int i = pixOffset; i < bitCount + pixOffset; i++)
                            dataArray[a + i] = raw[b + i];
                    }
                    else return null;
                }
            }
            return dataArray;
        }

        public static byte[] Swizzle(byte[] raw, int width, int height, int depth, int bitCount, bool deswizzle)
        {
            return Swizzle(raw, 0, width, height, depth, bitCount, deswizzle);
        }

        private static int Swizzle(int x, int y, int z, MaskSet masks)
        {
            return SwizzleAxis(x, masks.x) | SwizzleAxis(y, masks.y) | (z == -1 ? 0 : SwizzleAxis(z, masks.z));
        }

        private static int SwizzleAxis(int val, int mask)
        {
            int bit = 1;
            int result = 0;

            while (bit <= mask)
            {
                int tmp = mask & bit;

                if (tmp != 0) result |= (val & bit);
                else val <<= 1;

                bit <<= 1;
            }

            return result;
        }

        private class MaskSet
        {
            public int x = 0;
            public int y = 0;
            public int z = 0;

            public MaskSet(int w, int h, int d)
            {
                int bit = 1;
                int index = 1;

                while (bit < w || bit < h || bit < d)
                {
                    //if (bit == 0) break;
                    if (bit < w)
                    {
                        x |= index;
                        index <<= 1;
                    }
                    if (bit < h)
                    {
                        y |= index;
                        index <<= 1;
                    }
                    if (bit < d)
                    {
                        z |= index;
                        index <<= 1;
                    }
                    bit <<= 1;
                }
            }
        }
        #endregion

        public static ColorRGBA[] LoadHalo2Palette()
        {
            #region H2Palette

            byte[] H2PaletteBuffer = new byte[] {
                255, 126, 126, 255, 255, 127, 126, 255, 255, 128, 126, 255, 255, 129, 126, 255, 255, 126, 
                127, 255, 255, 127, 127, 255, 255, 128, 127, 255, 255, 129, 127, 255, 255, 126, 128, 255, 
                255, 127, 128, 255, 255, 128, 128, 255, 255, 129, 128, 255, 255, 126, 129, 255, 255, 127, 
                129, 255, 255, 128, 129, 255, 255, 129, 129, 255, 255, 130, 127, 255, 255, 127, 131, 255, 
                255, 127, 125, 255, 255, 131, 129, 255, 255, 124, 129, 255, 255, 130, 124, 255, 255, 129, 
                132, 255, 255, 124, 125, 255, 255, 133, 127, 255, 255, 125, 132, 255, 255, 128, 122, 255, 
                255, 132, 132, 255, 255, 122, 128, 255, 255, 133, 124, 255, 255, 127, 135, 255, 255, 124, 
                122, 255, 255, 136, 130, 255, 255, 121, 132, 255, 255, 131, 120, 255, 255, 132, 136, 255, 
                255, 119, 124, 255, 255, 137, 125, 255, 255, 123, 137, 255, 255, 125, 118, 255, 255, 137, 
                134, 255, 255, 117, 130, 255, 255, 135, 119, 255, 255, 129, 140, 255, 255, 119, 120, 255, 
                255, 141, 128, 255, 255, 119, 137, 255, 255, 129, 115, 255, 255, 136, 139, 255, 255, 114, 
                126, 255, 255, 140, 120, 255, 255, 124, 142, 255, 255, 121, 115, 255, 255, 142, 133, 255, 
                255, 113, 134, 255, 254, 135, 113, 255, 254, 133, 144, 255, 254, 113, 120, 255, 254, 145, 
                124, 255, 254, 118, 142, 255, 254, 126, 110, 255, 254, 142, 140, 255, 254, 109, 129, 255, 
                254, 142, 114, 255, 254, 127, 147, 255, 254, 115, 113, 255, 254, 148, 131, 255, 254, 111, 
                140, 255, 254, 133, 107, 255, 254, 139, 147, 255, 254, 107, 121, 255, 254, 148, 119, 255, 
                253, 119, 149, 255, 253, 120, 106, 255, 253, 149, 139, 255, 253, 105, 134, 255, 253, 141, 
                108, 255, 253, 132, 152, 255, 253, 108, 113, 255, 253, 153, 126, 255, 253, 111, 147, 255, 
                253, 128, 102, 255, 253, 146, 147, 255, 253, 101, 126, 255, 253, 150, 111, 255, 252, 123, 
                155, 255, 252, 113, 104, 255, 252, 155, 135, 255, 252, 103, 141, 255, 252, 138, 101, 255, 
                252, 139, 155, 255, 252, 101, 115, 255, 252, 157, 119, 255, 252, 113, 155, 255, 252, 121, 
                98, 255, 252, 154, 146, 255, 251, 96, 132, 255, 251, 149, 103, 255, 251, 129, 161, 255, 
                251, 105, 105, 255, 251, 161, 129, 255, 251, 102, 150, 255, 251, 132, 94, 255, 251, 148, 
                156, 255, 251, 94, 120, 255, 251, 159, 110, 255, 250, 117, 162, 255, 250, 113, 95, 255, 
                250, 162, 142, 255, 250, 93, 141, 255, 250, 145, 95, 255, 250, 138, 164, 255, 250, 96, 
                108, 255, 250, 166, 121, 255, 249, 104, 159, 255, 249, 125, 89, 255, 249, 157, 155, 255, 
                249, 88, 128, 255, 249, 158, 101, 255, 249, 124, 169, 255, 249, 103, 95, 255, 248, 169, 
                135, 255, 248, 92, 151, 255, 248, 139, 87, 255, 248, 148, 166, 255, 248, 87, 113, 255, 
                248, 168, 111, 255, 248, 109, 168, 255, 247, 115, 86, 255, 247, 167, 150, 255, 247, 84, 
                138, 255, 247, 154, 91, 255, 247, 134, 174, 255, 247, 92, 98, 255, 247, 175, 126, 255, 
                246, 94, 162, 255, 246, 130, 80, 255, 246, 159, 165, 255, 246, 80, 122, 255, 246, 168, 
                100, 255, 246, 117, 176, 255, 245, 103, 85, 255, 245, 176, 143, 255, 245, 82, 149, 255, 
                245, 148, 81, 255, 245, 146, 176, 255, 244, 82, 104, 255, 244, 178, 114, 255, 244, 100, 
                172, 255, 244, 119, 76, 255, 244, 170, 161, 255, 244, 74, 133, 255, 243, 165, 88, 255, 
                243, 128, 183, 255, 243, 91, 87, 255, 243, 183, 133, 255, 243, 84, 162, 255, 242, 138, 
                73, 255, 242, 158, 176, 255, 242, 73, 113, 255, 242, 179, 101, 255, 242, 108, 182, 255, 
                241, 106, 74, 255, 241, 181, 153, 255, 241, 72, 146, 255, 241, 158, 76, 255, 240, 141, 
                187, 255, 240, 79, 93, 255, 240, 188, 120, 255, 240, 89, 175, 255, 240, 125, 66, 255, 
                239, 172, 172, 255, 239, 66, 125, 255, 239, 176, 88, 255, 239, 120, 191, 255, 238, 92, 
                76, 255, 238, 191, 142, 255, 238, 72, 160, 255, 238, 148, 66, 255, 237, 156, 187, 255, 
                237, 67, 103, 255, 237, 190, 105, 255, 237, 97, 187, 255, 237, 111, 63, 255, 236, 185, 
                164, 255, 236, 61, 140, 255, 236, 170, 74, 255, 235, 134, 196, 255, 235, 77, 81, 255, 
                235, 197, 128, 255, 235, 77, 175, 255, 234, 134, 58, 255, 234, 171, 184, 255, 234, 58, 
                116, 255, 234, 188, 90, 255, 233, 109, 197, 255, 233, 95, 64, 255, 233, 196, 153, 255, 
                233, 61, 156, 255, 232, 159, 62, 255, 232, 150, 198, 255, 232, 64, 91, 255, 231, 201, 
                112, 255, 231, 85, 189, 255, 231, 118, 53, 255, 231, 186, 177, 255, 230, 52, 131, 255, 
                230, 182, 74, 255, 230, 125, 205, 255, 229, 78, 69, 255, 229, 205, 138, 255, 229, 64, 
                173, 255, 228, 145, 51, 255, 228, 167, 196, 255, 228, 52, 104, 255, 227, 200, 94, 255, 
                227, 97, 202, 255, 227, 101, 52, 255, 227, 200, 165, 255, 226, 49, 149, 255, 226, 172, 
                59, 255, 226, 142, 209, 255, 225, 63, 78, 255, 225, 211, 121, 255, 225, 72, 189, 255, 
                224, 128, 44, 255, 224, 185, 190, 255, 224, 44, 121, 255, 223, 195, 76, 255, 223, 113, 
                212, 255, 223, 82, 56, 255, 222, 211, 150, 255, 222, 51, 168, 255, 221, 158, 47, 255, 
                221, 161, 209, 255, 221, 49, 91, 255, 220, 212, 102, 255, 220, 84, 204, 255, 220, 109, 
                41, 255, 219, 201, 179, 255, 219, 39, 140, 255, 219, 186, 59, 255, 218, 132, 218, 255, 
                218, 64, 64, 255, 217, 219, 132, 255, 217, 58, 187, 255, 217, 140, 37, 255, 216, 181, 
                203, 255, 216, 38, 108, 255, 216, 208, 82, 255, 215, 100, 217, 255, 215, 89, 43, 255, 
                214, 215, 164, 255, 214, 39, 160, 255, 214, 172, 44, 255, 255, 128, 128, 0, };

            #endregion

            ColorRGBA[] H2Palette = new ColorRGBA[256];
            for (int x = 0; x < 256; x++)
            {
                int r = x * 4;
                H2Palette[x].r = H2PaletteBuffer[r];
                H2Palette[x].g = H2PaletteBuffer[r + 1];
                H2Palette[x].b = H2PaletteBuffer[r + 2];
                H2Palette[x].a = H2PaletteBuffer[r + 3];
            }
            return H2Palette;
        }
    }
}
