﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using HaloMap.Decorators;

namespace HaloMap.Decorators
{
    #region Decorator Raw Layout
    public class ResourceBlock
    {
        #region Fields
        public short ChunkSize, ChunkCount;
        public int Offset;
        public int Size;
        #endregion

        #region Methods
        public void Read(Map map)
        {
            map.br.BaseStream.Position += 8;
            Size = map.br.ReadInt32();
            Offset = map.br.ReadInt32();
        }
        #endregion
    }

    public class Vertice
    {
        #region Fields
        public float X, Y, Z, NormalI, NormalJ, NormalK, TangentI, TangentJ, TangentK, BinormalI, BinormalJ, BitnormalK, U, V;
        #endregion
    }
    public class VerticeBlock : ResourceBlock
    {
        #region Fields
        public List<Vertice> Verts = new List<Vertice>();
        #endregion

        #region Constructor
        public VerticeBlock(Map map)
        {
            base.Read(map);
        }
        #endregion

        #region Methods
        public void Read(BinaryReader br, int Offset)
        {
            br.BaseStream.Position = Offset + base.Offset;
            for (int i = 0; i < base.Size / 56; i++)
            {
                Vertice Vert = new Vertice();
                Vert.X = br.ReadSingle();
                Vert.Y = br.ReadSingle();
                Vert.Z = br.ReadSingle();
                Vert.NormalI = br.ReadSingle();
                Vert.NormalJ = br.ReadSingle();
                Vert.NormalK = br.ReadSingle();
                Vert.TangentI = br.ReadSingle();
                Vert.TangentJ = br.ReadSingle();
                Vert.TangentK = br.ReadSingle();
                Vert.BinormalI = br.ReadSingle();
                Vert.BinormalJ = br.ReadSingle();
                Vert.BitnormalK = br.ReadSingle();
                Vert.U = br.ReadSingle();
                Vert.V = br.ReadSingle(); ;
                Verts.Add(Vert);
            }
        }
        #endregion
    }
    #endregion
}

namespace HaloMap
{
    public class Decorator
    {
        #region Fields
        public int Offset, Size, HeaderSize, DataSize;
        public RawLocation Location;
        public VerticeBlock Verts;
        public TagReference Shad = new TagReference();
        public Shader Shader;
        #endregion

        #region Constructor
        public Decorator(Map map)
        {
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset;
            Reflexive Shads = new Reflexive();
            Shads.Read(map, true);
            map.br.BaseStream.Position = Shads.Translation;
            Shad.Read(map, true);
            Shader = new Shader(map, Shad.Tag);
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 56;
            Offset = map.br.ReadInt32();
            Size = map.br.ReadInt32();
            HeaderSize = map.br.ReadInt32();
            DataSize = map.br.ReadInt32();
            Location = RawData.RawLocator(ref Offset);
            Reflexive Resources = new Reflexive();
            Resources.Read(map, true);
            if (Resources.ChunkCount > 0)
            {
                map.br.BaseStream.Position = Resources.Translation;
                Verts = new VerticeBlock(map);
                Verts.Read(map.OpenMap(Location), Offset + 40);
            }
        }
        #endregion
    }
}
